<?php

namespace FourWords\AppManager\Appstore\Livewire\Admin;

use FourWords\AppManager\Appstore\Models\AppstoreType;
use FourWords\AppManager\Ioc\Livewire\IocComponent;
use FourWords\AppManager\Appstore\Models\AppstoreApp;
use FourWords\AppManager\Appstore\Models\AppstoreVersion;
use Maatwebsite\Excel\Facades\Excel;
use FourWords\AppManager\Appstore\Imports\AbmCodesImport;
use Illuminate\Validation\ValidationException;
use Livewire\WithFileUploads;

class CreateApp extends IocComponent
{
    use WithFileUploads;

    private $appExtension;
    private $successMsg = [];
    private $errorMsg = [];

    public $appInfo = [
        'name' => '',
        'bundleID' => '',
        'os' => '1',
        'desc' => '',
        'logo' => '',
        'abm' => '',
        'abmActive' => false
    ];
    public $versionInfo = [
        'version' => '1.0.0',
        'file' => '',
        'dimension' => '',
        'changelog' => ''
    ];

    protected function rules() {
        $rules = [
            'appInfo.name' => 'required',
            'appInfo.bundleID' => 'required',
            'versionInfo.version' => 'nullable|regex:/^[0-9.]+$/',
        ];

        if(!$this->appInfo['abmActive']) {            
            $rules += ['versionInfo.file' => 'required'];
        }

        return $rules;
    }

    protected $validationAttributes = [
        'appInfo.name' => 'name',
        'appInfo.bundleID' => 'bundle identifier',
        'versionInfo.version' => 'version',
        'versionInfo.file' => 'file'
    ];

    public function updated($name) {
        if($name == 'appInfo.os') {
            if($this->appInfo['os'] != 1) {
                $this->appInfo['abmActive'] = false;
            }
        }

        if($name == 'appInfo.logo') {
            if(!in_array($this->appInfo['logo']->extension(), ['jpg', 'jpeg', 'png', 'gif', 'tiff', 'svg'])) {
                $this->errorToast('Il logo deve essere un\'immagine');
                $this->appInfo['logo'] = null;
            }
        }
    }

    public function render() {
        $appType = AppstoreType::all();
        return view('appstore::admin.livewire.create-app', ['appType' => $appType]);
    }

    public function createApp() {
        $this->validate();
        if($this->checkImage()) {
            try {
                if($this->insertDB()) {
                    array_push($this->successMsg, 'App creata con successo');
                    return redirect(route('appstore.admin.index'))->with(['successMsgFromRedirect' => $this->successMsg, 'errorMsgFromRedirect' => $this->errorMsg]);
                }
            } catch (\Exception $e) {
                $this->errorToast($e->getMessage());
            }
        }
        return null;
    }

    private function appDB() {
        $app = new AppstoreApp();

        //$app->id = AppstoreApp::all()->last()->id + 1;
        $app->name = $this->appInfo['name'];
        $app->bundle_identifier = $this->appInfo['bundleID'];
        $app->type = $this->appInfo['os'];
        $app->description = $this->appInfo['desc'];
        $app->save();
        $app->icon = $this->uploadImage($app);
        $app->save();

        $app->abm_active = $this->appInfo['abmActive'];
        if($this->checkAbm($app)) {
            $this->uploadAbm($app);
        } else if($this->appInfo['abmActive']) {
            array_push($this->errorMsg, 'Nessun codice ABM rilevato, l\'app sarà disattivata');
            $app->active = false;
        }

        $app->save();

        return $app;
    }

    private function versionDB($app) {
        if($this->versionInfo['version'] != null and $app != null) {
            try {
                $version = new AppstoreVersion();
                $version->app = $app->id;
                $version->version = $this->versionInfo['version'];
                
                $version->changelog = $this->versionInfo['changelog'];
                $version->show_in_history = true;
                $version->active = true;
                
                return $version;
            } catch(\Exception $e) {
                return null;
            }
        }
        return null;
    }

    private function uploadImage($app) {
        if($this->appInfo['logo'] != null) {
            $this->appInfo['logo']->storeAs('logo','logo-app'.$app->id.'.'.$this->appInfo['logo']->extension(), 'public');
            //$this->appInfo['currentLogo'] = 'logo-'.$this->app->name.'.jpg';
            return 'logo-app'.$app->id.'.'.$this->appInfo['logo']->extension();
        }
        return null;
    }

    private function checkImage() {
        try {
            $this->validate(['appInfo.logo' => 'max:30720']); //Controllo logo
            if($this->appInfo['logo'] != null) {
                if(!in_array($this->appInfo['logo']->extension(), config('appstore.admin.imageFileExtensions'))) {
                    $this->errorToast('Il logo deve essere un\'immagine');
                    return false;
                }
            }
        } catch (\Exception $e) {
            $this->errorToast('Il logo deve pesare meno di 30MB');
            return false;
        }
        return true;
    }

    private function insertDB() {
        try {
            if(!$this->checkFile() && !$this->appInfo['abmActive']) {
                return false;
            }

            $app = $this->appDB();
            if($app == null) {
                return false;
            }

            $version = $this->versionDB($app);
            if($version != null) {
                $version->save();

                if(!$this->appInfo['abmActive']) {
                    $this->uploadFile($app, $version);
                }
            }
        } catch(\Throwable $e) {
            $this->errorToast($e->getMessage());
            //$app->delete();
            return false;
        }
        return true;
    }

    private function checkAbm() {
        if($this->appInfo['os'] != 1) {
            return false;
        }

        if(!$this->appInfo['abmActive']) {
            return false;
        }

        if($this->appInfo['abm'] == null) {
            return false;
        }
        
        if(!in_array($this->appInfo['abm']->extension(), [config('appstore.admin.abmFileExtensions')])) {
            $this->errorToast('Il file dell\'ABM deve essere di tipo .'.config('appstore.admin.abmFileExtensions'));
            return false;
        }

        return true;
    }

    private function uploadAbm($app) {
        if($this->appInfo['abm'] != null) {
            $abmImport = new AbmCodesImport($app->id);
            Excel::import($abmImport, 'livewire-tmp/'.$this->appInfo['abm']->getfilename(), 'local');
            
            if($abmImport->countAbmCodes() > 0) {
                $this->successToast('Sono stati registrati '.$abmImport->countAbmCodes().' codici ABM');
            } else {
                $this->errorToast('Nessun codice ABM rilevato, l\'app sarà disattivata');
                if($this->countAbmCodes() == 0) {
                    $app->active = false;
                    $app->abm_active = false;
                    $this->appInfo['abmActive'] = false;
                }
            }
            $app->save();
        }
    }

    private function checkFile() {
        if($this->versionInfo['file'] == null) {
            return false;
        }

        $this->appExtension = AppstoreType::find($this->appInfo['os'])->extension;
        $fileNameToGetExtension = explode('.',$this->versionInfo['file']->getClientOriginalName());
        if(!in_array($fileNameToGetExtension[count($fileNameToGetExtension) - 1], [$this->appExtension])) {
            $this->errorToast('Il file dell\'app deve essere di tipo .'.$this->appExtension);
            return false;
        }

        return true;
    }

    private function uploadFile($app, $appVersion) {
        $filename = $app->name.'_'.$appVersion->version.'.'.$this->appExtension;

        $this->versionInfo['file']->storeAs('file', $filename, 'public');
        $appVersion->file = $filename;

        $appVersion->dimension = $this->fileSizeConvertMB($this->versionInfo['file']->getSize());
        $appVersion->save();
    }

    private function fileSizeConvertMB($bytes) {
        return number_format($bytes / 1048576, 2) . ' MB';
    }
}
