<?php

namespace FourWords\AppManager\Appstore\Livewire\Admin;

use FourWords\AppManager\Appstore\Models\AbmCode;
use FourWords\AppManager\Appstore\Models\AppstoreType;
use FourWords\AppManager\Ioc\Livewire\IocComponent;
use FourWords\AppManager\Appstore\Models\AppstoreVersion;
use Livewire\WithFileUploads;
use Maatwebsite\Excel\Facades\Excel;
use FourWords\AppManager\Appstore\Imports\AbmCodesImport;

class FormAppEdit extends IocComponent {
    use WithFileUploads;

    public $randId = 1;

    public $app;
    public $appInfo = [
        'name' => '',
        'bundleID' => '',
        'os' => '',
        'desc' => '',
        'logo' => '',
        'showLogo' => '',
        'abm' => '',
        'abmCheck' => '',
        'active' => ''
    ];

    protected $rules = [
        'appInfo.name' => 'required',
        'appInfo.bundleID' => 'required'
    ];
    protected $validationAttributes = [
        'appInfo.name' => 'name',
        'appInfo.bundleID' => 'bundle identifier'
    ];
    protected $listeners = ['toggleAppStatus'];

    public function mount() {
        $this->appInfo['name'] = $this->app->name;
        $this->appInfo['bundleID'] = $this->app->bundle_identifier;
        $this->appInfo['os'] = $this->app->type;
        $this->appInfo['desc'] = $this->app->description;
        $this->appInfo['showLogo'] = $this->app->icon.'?'.time();
        $this->appInfo['abmActive'] = $this->app->abm_active;
        $this->appInfo['active'] = $this->app->active;
    }

    public function updated($name) {
        if($name == 'appInfo.os') {
            if($this->appInfo['os'] != 1) {
                $this->appInfo['abmActive'] = false;
            }
        }

        if($name == 'appInfo.logo') {
            if(!in_array($this->appInfo['logo']->extension(), ['jpg', 'jpeg', 'png', 'gif', 'tiff', 'svg'])) {
                $this->errorToast('Il logo deve essere un\'immagine');
                $this->appInfo['logo'] = null;
            }
        }
    }

    public function render() {
        $appType = AppstoreType::all();
        return view('appstore::admin.livewire.form-app-edit', ['appType' => $appType]);
    }

    public function toggleAppStatus() {
        if($this->app->abm_active && AbmCode::countAbmCodes($this->app->id) == 0 && !$this->app->active) {
            $this->errorToast('Nessun codice ABM rilevato, per attivare l\'app è necessario caricare codici ABM o disattivare la funzione ABM');
            $this->emit('setAppActive', false);
            return ;
        }

        $this->app->active = !$this->app->active;
        $this->app->save();
    }

    public function update() {
        if(!$this->checkPermission('edit-appstore')) {
            return;
        }

        if($this->checkUpdate()) {
            try {
                $this->appDB();
                $this->uploadImage();

                $this->successToast('App aggiornata con successo');
                $this->emit('refreshApp');
                $this->emit('refreshVersion');
            } catch (\Exception $e) {
                $this->errorToast($e->getMessage());
            }
        }

        $this->refreshUploadFile();
    }

    private function checkUpdate() {
        $this->validate();
        if(!$this->checkImage()) {
            return false;
        }

        if($this->appInfo['abm'] == null && $this->appInfo['abmActive'] && AbmCode::countAbmCodes($this->app->id) == 0 && $this->app->active) {
            $this->errorToast('Nessun codice ABM rilevato, per procedere è necessario disattivare l\'app');
            return false;
        }
        
        return true;
    }

    public function delete() {
        AppstoreVersion::where("app", "=", $this->app->id)->delete();
        $this->app->delete();
        //$this->successToast('App eliminata con successo');
        return redirect(route('appstore.admin.index'))->with('successMsgFromRedirect', ['App eliminata con successo']);
    }

    private function appDB() {
        $this->app->name = $this->appInfo['name'];
        $this->app->bundle_identifier = $this->appInfo['bundleID'];
        $this->app->type = $this->appInfo['os'];
        $this->app->description = $this->appInfo['desc'];

        $this->app->abm_active = $this->appInfo['abmActive'];
        if($this->app->abm_active && $this->checkAbm()) {
            $this->uploadAbm();
        }
        $this->app->save();
    }

    private function uploadImage() {
        if(!$this->checkPermission('edit-appstore')) {
            return;
        }

        if($this->appInfo['logo'] != null) {
            $this->appInfo['logo']->storeAs('logo','logo-app'.$this->app->id.'.'.$this->appInfo['logo']->extension(), 'public');
            $this->app->icon = 'logo-app'.$this->app->id.'.'.$this->appInfo['logo']->extension();
            $this->app->save();

            //$this->removeImage($this->appInfo['showLogo']);
            $this->appInfo['showLogo'] = $this->app->icon.'?'.time();
        }
    }

    private function checkImage() {
        try {
            $this->validate(['appInfo.logo' => 'max:30720']); //Controllo logo
            if($this->appInfo['logo'] != null) {
                if(!in_array($this->appInfo['logo']->extension(), config('appstore.admin.imageFileExtensions'))) {
                    $this->errorToast('Il logo deve essere un\'immagine');
                    return false;
                }
            }
        } catch (\Exception $e) {
            $this->errorToast('Il logo deve pesare meno di 30MB');
            return false;
        }
        return true;
    }

    private function showNewImage() {
        $this->appInfo['logo']->storeAs('logo','demo-logo-app'.$this->app->id.'.'.$this->appInfo['logo']->extension(), 'public');
        $this->appInfo['showLogo'] = 'demo-logo-app'.$this->app->id.'.'.$this->appInfo['logo']->extension();
    }

    private function checkAbm() {
        if($this->appInfo['os'] != 1) {
            return false;
        }

        if(!$this->appInfo['abmActive']) {
            return false;
        }

        if($this->appInfo['abm'] == null) {
            if($this->app->active && AbmCode::countAbmCodes($this->app->id) == 0) {
                
            }
        } else if(!in_array($this->appInfo['abm']->extension(), config('appstore.admin.abmFileExtensions'))) {
            $this->errorToast('Il file dell\'ABM deve essere di tipo .'.config('appstore.admin.abmFileExtensions'));
            return false;
        }

        return true;
    }

    private function uploadAbm() {
        if($this->appInfo['abm'] != null) {
            $abmImport = new AbmCodesImport($this->app->id);
            Excel::import($abmImport, 'livewire-tmp/'.$this->appInfo['abm']->getfilename(), 'local');

            if(session()->has('abmCodeErrorLength')) {
                $this->errorToast(session()->get('abmCodeErrorLength'));
            }
            
            if($abmImport->countAbmCodes() > 0) {
                $this->successToast('Sono stati registrati '.$abmImport->countAbmCodes().' codici ABM');
            } else {
                $this->errorToast('Nessun codice ABM è stato registrato');
                if(AbmCode::countAbmCodes($this->app->id) == 0 && $this->app->active) {
                    $this->app->abm_active = false;
                    $this->appInfo['abmActive'] = false;
                }
            }
        }
    }

    private function refreshUploadFile() {
        $this->appInfo['abm'] = '';
        $this->appInfo['logo'] = '';
        //Per svuotare l'input file serve modificare il suo ID
        //Assegno $randId e lo incremento ogni volta che devo svuotare l'input
        $this->randId++; 
    }
}
