<?php

namespace FourWords\AppManager\Catalog\Importer;

use DateTimeZone;
use Exception;
use File;
use FourWords\AppManager\Catalog\Resources\ResourceAttributes;
use Illuminate\Support\Facades\DB;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Logger;
use SplFileInfo;

abstract class AbstractFileImporter
{

    /** @var Logger */
    public $logger;

    abstract protected function getLoggerName(): string;
    abstract protected function getDestinationPath(): string;
    abstract protected function getAllowedExtensions(): array;
    abstract protected function importFiles();

    public function __construct()
    {
        $this->initializeLogger();
    }

    public function pushHandler($handler) {
        $this->logger->pushHandler($handler);
    }

    protected function getFileRegex() : string
    {
        return '';
    }

    protected function generateImportedFilename(SplFileInfo $file, string $additionalAttributes = ''): string {
        // md5(nome.estensione+dimensione).estensione
        return md5($file->getFilename() . $file->getSize() . $additionalAttributes).".".$file->getExtension();
    }

    public function import()
    {

        $this->createDestinationDirectory($this->getDestinationPath());

        try {
            DB::beginTransaction();
            $this->importFiles();
            DB::commit();
        } catch (Exception $e) {
            DB::rollBack();
            throw $e;
        }
    }


    protected function initializeLogger()
    {
        $this->logger = new Logger($this->getLoggerName());
        $this->logger->setTimezone(new DateTimeZone('Europe/Rome'));

        if (!file_exists(config('catalog.log.dir'))) {
            mkdir(config('catalog.log.dir'), config('catalog.log.dirPermissions'), true);
        }

        $logHandler = new RotatingFileHandler(config('catalog.log.dir') . 'debug.log', 14, Logger::DEBUG, true, config('catalog.log.filePermissions'));
        $logDateFormat = "Y-m-d H:i:s";
        $logOutput = "[%datetime%] %channel%.%level_name% : %message% \n";
        $logFormatter = new LineFormatter($logOutput, $logDateFormat, true);
        $logHandler->setFormatter($logFormatter);
        $this->logger->pushHandler($logHandler);
    }

    protected function createDestinationDirectory(string $directoryPath)
    {
        if (!is_dir($directoryPath) && !mkdir($directoryPath, 0775, true)) {
            throw new Exception("Impossibile creare directory di destinazione: $directoryPath");
        }
    }

    protected function validateFileSizeInMegaBytes(SplFileInfo $file, int $maxSize)
    {
        if ($file->getSize() > $maxSize * 1024 * 1024) {
            $this->logger->warning('File con dimensioni troppo grandi: ' . ($file->getSize() / 1024 / 1024) . '/' . $maxSize . ' (MB): ' . $file->getPathname());
            return false;
        }
        return true;
    }

    protected function copyFile(SplFileInfo $sourceFile, string $destinationPath) : bool
    {
        if (!File::exists($destinationPath)) {
            if (!File::copy($sourceFile->getPathname(), $destinationPath)) {
                $this->logger->error('Impossibile copiare il file '.$sourceFile->getFilename().' in: ' . $destinationPath);
            } else {
                $this->logger->debug("Copiato file " . $sourceFile->getFilename() . " in: " . $destinationPath);
                return true;
            }
        }
        return false;
    }
}