<?php

namespace FourWords\AppManager\Catalog\Importer;

use Exception;
use FourWords\AppManager\DataExtractor\DataExtractor;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\Facades\DB;

abstract class AbstractImporter implements ImporterInterface
{

    protected $dataExtractor;
    protected $model;
    protected $useMutators = false;

    protected $underTransaction = false;

    /** @var bool Lancio un'exception se non vengono restituiti dei risultati per popolare la tabella */
    protected $throwIfEmptyResult = false;

    public function __construct(Model $model, DataExtractor $dataExtractor)
    {
        ini_set('memory_limit', '8192M');
        set_time_limit(60*60*8);

        $this->model = $model;
        $this->dataExtractor = $dataExtractor;
        $this->dataExtractor->setConfig(config('database.connections.catalog'));
    }

    protected function onInsertRows(&$rows) {
        return $rows;
    }

    protected function onInsertRecord(&$record) {
        return $record;
    }

    protected function truncate($model = null) {

        if (is_string($model))
            $model = new $model();

        if ($model == null)
            $model = $this->model;

        $model->getConnection()->statement('SET FOREIGN_KEY_CHECKS = 0;');
        DB::table($model->getTable())->truncate();
        $model->getConnection()->statement('SET FOREIGN_KEY_CHECKS = 1;');
    }

    public function import() {

        try {

            if ($this->underTransaction) DB::beginTransaction();

            $this->truncate();
            $records = $this->getRecords();
            if ($this->throwIfEmptyResult && count($records) == 0) {
                $className = basename(str_replace('\\', '/', trim(get_class($this), '\\')));
                throw new Exception("$className :: Il risultato restituito è vuoto quindi fermo l'importazione");
            }
            $this->insert($records);

            if ($this->underTransaction) DB::commit();

        } catch (Exception $e) {
            if ($this->underTransaction) DB::rollBack();
            throw $e;
        }


    }

    protected function insert($rows, $model = null) {

        if (is_string($model))
            $model = new $model();

        if (!$model)
            $model = $this->model;

        if (count($rows) == 0) {
            return;
        }

        $tableColumns = $model->getConnection()->getSchemaBuilder()->getColumnListing($model->getTable());

        foreach ($rows as $key => $record) {
            $intersection = array_intersect_key($record, array_flip($tableColumns));
            if (!empty($intersection)) $rows[$key] = $intersection;
        }

        $this->onInsertRows($rows);

        if ($this->useMutators) {
            foreach ($rows as $row) {

                // Trasformo le stringhe vuote '' in null
                $record = FWUtils::removeNullsAndEmptyString($row);

                $this->onInsertRecord($record);
                $instance = new $model($record);
                $instance->save();
            }
        } else {

            $count = 0;
            $split = 10000;
            $start = 0;

            while ($count < count($rows)) {

                $values = [];

                $previousRecord = null;

                for ($index = $start; $index < count($rows); $index++) {

                    if ($rows[$index] == null) continue;

                    $rows[$index] = (array)$rows[$index];
                    $this->onInsertRecord($rows[$index]);


                    $rowWithoutNulls = FWUtils::removeNullsAndEmptyString($rows[$index]);

                    if ($previousRecord && (count($previousRecord) != count($rowWithoutNulls)) || array_diff_key($previousRecord, $rowWithoutNulls)) {

                        // Se il numero di colonne è diverso dal numero di colonne del record precedente, devo spezzare qui e ripartire da questo indice
                        $start = $index;
                        break;
                    }

                    $previousRecord = $rowWithoutNulls;

                    $rows[$index] = FWUtils::inQuotes($rowWithoutNulls);

                    $values[] = "(" . implode(",", array_values($rows[$index])) . ")";
                    $count += 1;

                    if (count($values) == $split) {
                        $start = $index+1;
                        break;
                    }
                }

                $insert = "INSERT INTO ".$model->getTable()." (".implode(",",array_keys( $previousRecord)).") VALUES ".implode(",",$values);
                $model->getConnection()->unprepared($insert);
            }

        }

    }

}