<?php


namespace FourWords\AppManager\Catalog\Importer;

use Exception;
use FourWords\AppManager\Catalog\Model\AttributeValue;
use FourWords\AppManager\Catalog\Model\CustomerFeature;
use FourWords\AppManager\Catalog\Model\CustomerHasFeature;
use FourWords\AppManager\DataExtractor\DataExtractor;
use FourWords\AppManager\Ioc\Models\IocUser;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;

class CustomerFeatureImporter extends SyncableAbstractImporter
{

    protected $throwIfEmptyResult = true;
    protected $featureIdsByCode = [];
    /** @var Collection */
    protected $productIdsByCode = [];

    protected $relationModel;

    public function __construct(CustomerFeature $model, CustomerHasFeature $relationModel, DataExtractor $dataExtractor)
    {
        parent::__construct($model, $dataExtractor);
        $this->relationModel = $relationModel;
    }

    /** Mappatura Nome Feature => Nome alias nella query */
    function getCustomFeature() {
        return [
            'Classifica 1'              => 'cla1ca',
            'Classifica 2'              => 'cla2ca',
            'Classifica 3'              => 'cla3ca',
            'Classifica 4'              => 'cla4ca',
            'Classifica 5'              => 'cla5ca',
        ];
    }

    public function getRecords()
    {
        $lib = config('catalog.as.library');
        $query = "SELECT DISTINCT
    CONTCA AS customer_code,
    CLA1CA AS cla1ca,
    CLA2CA AS cla2ca,
    CLA3CA AS cla3ca,
    CLA4CA AS cla4ca,
    CLA5CA AS cla5ca
     FROM ".$lib."CGANA00F";

        $customerCodesAndCodedFeatuers = $this->dataExtractor->execute($query);
        $this->decodeFeatureValue($customerCodesAndCodedFeatuers);
        return $customerCodesAndCodedFeatuers;

    }

    public function decodeFeatureValue(&$features)
    {
        // Sostituisco le codifiche delle classifiche con il valore
        $classifiche = array_values($this->getCustomFeature());
        $attributeValues = AttributeValue::whereIn('classifica', $classifiche)->get();
        foreach ($features as &$item) {
            foreach ($classifiche as $classifica) {

                $codifica = $item[$classifica];

                if (isset($item[$classifica])) {
                    $attributeValue = $attributeValues->where('classifica', $classifica)->where('codifica', $codifica)->first();

                    if ($attributeValue) {
                        $item[$classifica] = $attributeValue->valore;
                    }
                }

            }
        }
    }

    public function import()
    {

        try {

            DB::beginTransaction();

            $this->model->setAllIsSynchronized(false);
            $this->truncate($this->relationModel);

            // Creo le Feature specificate in getCustomFeature()
            $this->createFeatures();

            $this->connectFeaturesToCustomers();

            $this->model->deleteNotSynchronized();

            DB::commit();

        } catch (Exception $e) {
            DB::rollBack();
            throw $e;
        }

    }

    protected function createFeatures() {
        foreach ($this->getCustomFeature() as $featureName => $featureCode ) {
            // Creo la Feature
            $feature = $this->model->sync(['code' => $featureCode, 'name' => $featureName]);

            // Salvo l'id con chiave 'code', che riutilizzerò per collegare i valori della feature al cliente
            $this->featureIdsByCode[$featureCode] = $feature->id;
        }
    }

    protected function connectFeaturesToCustomers() {
        // Prendo codici e id cliente importati precedentemente
        $this->customerIdsByCode = IocUser::pluck('id','code');

        // Prendo il risultato della query
        $customerAndFeaturesValues = $this->getRecords();

        // Aggiungo ai codici cliente i relativi id, ottenuti importando i prodotti nel DB Laravel
        $customerIdsByCode = $this->customerIdsByCode;
        $customerAndFeaturesValues = array_map(function ($row) use ($customerIdsByCode) {
            $row['id'] = $customerIdsByCode[$row['customer_code']];
            return $row;
        }, $customerAndFeaturesValues);

        // Per ogni record con codice cliente e valori della feature
        foreach ($customerAndFeaturesValues as $customer) {

            if (!$customer['id']) continue;

            // Per ogni valore feature
            foreach ($this->featureIdsByCode as $featureCode => $featureId) {
                // Collego il cliente alla feature con il relativo valore
                if (!empty($customer[$featureCode])) {
                    $this->relationModel->create(['feature_id' => $featureId, 'customer_id' => $customer['id'], 'value' => $customer[$featureCode]]);
                }

            }
        }
    }
}
