<?php


namespace FourWords\AppManager\Catalog\Importer;

use Exception;
use FourWords\AppManager\Catalog\Model\ImageGallery;
use FourWords\AppManager\Catalog\Model\ProductImage;
use FourWords\AppManager\Catalog\Model\SyncableModel;
use FourWords\AppManager\Catalog\Resources\ResourceAttributes;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Support\Facades\File;
use Intervention\Image\ImageManagerStatic as ImageResizer;
use SplFileInfo;

class ImageImporter extends AbstractFileImporter
{
    /** @var SyncableModel */
    protected $productImageModel;
    /** @var SyncableModel */
    protected $imageGalleryModel;

    public function __construct(ProductImage $productImageModel, ImageGallery $imageGalleryModel)
    {
        parent::__construct();
        $this->productImageModel = $productImageModel;
        $this->imageGalleryModel = $imageGalleryModel;
    }

    protected function getLoggerName(): string
    {
        return 'catalog-images';
    }

    protected function getDestinationPath(): string
    {
        return config('catalog.images.dir');
    }

    /**
     * @throws \Throwable
     */
    public function importFiles() {

        // Imposto le immagini esistenti nel db come "non sincronizzate",
        // nel momento in cui importerò le immagini, quelle che ho reimportato e quindi "sincronizzate" rimarranno e le "non sincronizzate" verranno eliminate
        $this->temporarilySetImagesAsNotSynchronized();

        // Importo le immagini come "sincronizzate"
        $this->importImages(config('catalog.images.importDir'), true, -1);

        $className = basename(str_replace('\\', '/', trim(get_class($this), '\\')));
        if ($errorMessage = $this->validateSynchronization()) {
            throw new Exception("$className :: validateSynchronization() ->" . $errorMessage);
        }

        // Elimino le immagini che non sono più "sincronizzate"
        $this->deleteImagesNotSynchronized();
    }

    /**
     * Funzione per validare la sincronizzazione, la utilizzo ad esempio per evitare che le tabelle vengano totalmente (o quasi) svuotate se non trovo dei record per procedure lanciate da clienti sbarazzini o errori su servizi
     * @return false|string
     */
    protected function validateSynchronization()
    {
        if ($this->productImageModel->getNotSynchronizedPercentage() > 50) {
            return "La percentuale di file non sincronizzati supera il 50%";
        }

        return false;
    }

    protected function temporarilySetImagesAsNotSynchronized()
    {
        $this->imageGalleryModel->setAllIsSynchronized(false);
        $this->productImageModel->setAllIsSynchronized(false);
    }

    protected function deleteImagesNotSynchronized()
    {
        $this->imageGalleryModel->deleteNotSynchronized();
        $this->productImageModel->deleteNotSynchronized();
    }

    /**
     * @param SplFileInfo $file
     * @return ResourceAttributes
     * @throws Exception
     */
    protected function extractImageData(SplFileInfo $file): ResourceAttributes
    {

        $filename = $file->getFilename();
        $nameOnly = $file->getBasename(".".$file->getExtension());

        list($prefix, $code, $position) = explode("_", $nameOnly);

        // Se non c'è il codice lancio un'exception
        if (!$code) throw new Exception("Impossibile recuperare il codice dal file:".$filename);

        // Se non c'è il progressivo è la foto principale
        $pos = intval($position) ?? 0;

        $imageAttribute = new ResourceAttributes($filename, $file->getSize(), $pos+1);
        $imageAttribute->setCode($code);

        return $imageAttribute;
    }

    protected function getAllowedExtensions(): array
    {
        return ['jpg','jpeg','png','gif'];
    }

    protected function getFileRegex(): string
    {
        return '/^dett_\S+\.\w+$/';
    }

    /**
     * @param SplFileInfo $file
     * @return void
     */
    protected function resizeImage($file)
    {
        $img = ImageResizer::make($file);
        $originalWidth = $img->getWidth();
        $originalHeight = $img->getHeight();

        $desiredWidth = config('catalog.images.products.desiredWidth');
        $desiredHeight = config('catalog.images.products.desiredHeight');
        $desiredFormat = config('catalog.images.desiredFormat');

        // Se non ho impostato la qualità, ridimensiono solo se necessario
        if ($originalWidth <= $desiredWidth && $originalHeight <= $desiredHeight) {
            return;
        }

        $img->resize($desiredWidth, $desiredHeight, function ($constraint) {
            $constraint->aspectRatio(); // Mantengo le proporzioni
            $constraint->upsize(false); // Evito l'upsizing se una foto è più piccola delle dimensioni desiderate
        });

        $filenameWithoutExtension = $file->getBasename(".".$file->getExtension());

        $img->save(
            FWUtils::addTrailingSlashes(config('catalog.images.dir')).$filenameWithoutExtension.'.'.(!empty($desiredFormat) ? $desiredFormat : $file->getExtension()),
            config('catalog.images.products.desiredQuality'),
            $desiredFormat);

        $this->logger->debug("Ridimensionata immagine ".$file->getPathname());
    }

    protected function importImages($importDir, $recursively = true, $maxLevel = 2, $level = 1,  $galleryName = '') {

        $files = File::files($importDir);
        $allFiles = $recursively ? array_merge(File::directories($importDir),$files) : $files;

        // Se la cartella specificata ha file o cartelle creo la collection, altrimenti non faccio nulla
        if (count($allFiles) == 0) {
            $this->logger->debug("Cartella ignorata perché non contiene file: ".basename($importDir));
            return;
        }

        // Se il nome galleria non è specificato, lo prendo dalla cartella che contiene il file
        if (empty($galleryName)) {
            $galleryName = basename($importDir);
        }
        $galleryName = ucfirst($galleryName);

        // Se la galleria non esiste la creo, altrimenti imposto il flag synchronized a true
        /** @var SyncableModel $gallery */
        $gallery = $this->imageGalleryModel->sync(['name' => $galleryName]);

        if ($gallery->isNew()) {
            $this->logger->info("Nuova galleria: ".$galleryName);
        }

        foreach ($allFiles as $file) {

            if (File::isDirectory($file)) {

                if ($maxLevel == -1 || $level < $maxLevel) $this->importImages($file, $recursively, $maxLevel, $level+1);
                continue;

            } else if ($file->isFile()) {

                $fileMatchRegularExpression = $this->getFileRegex();
                if ( !empty($fileMatchRegularExpression) && !preg_match($fileMatchRegularExpression,$file->getFilename()) ) {
                    continue;
                }

                $allowedExtensions = $this->getAllowedExtensions();
                if (!empty($allowedExtensions) && !in_array($file->getExtension(), $allowedExtensions)) {
                    continue;
                }
            }

            if (!$this->validateFileSizeInMegaBytes($file, config('catalog.images.maxSize'))) {
                continue;
            }

            try {
                $imageAttributes = $this->extractImageData($file);
                $imageAttributes->setCollection($gallery->id);
            } catch (Exception $e) {
                $this->logger->warning('Immagine non importata, impossibile recuperare i dettagli del file '.$file->getPathname());
                continue;
            }

            // Se il record non esiste lo creo, altrimenti aggiorno i dati. Imposto il flag synchronized a true
            /** @var SyncableModel $productImage */
            $productImage = $this->productImageModel->sync(
                ['size' => $imageAttributes->getSize(), 'filename' => $imageAttributes->getFilename(), 'gallery' => $gallery->id],
                [
                    'product_code' => $imageAttributes->getCode(),
                    'position' => $imageAttributes->getPosition(),
                    'gallery' => $gallery->id,
                    'imported_filename' => $this->generateImportedFilename($file)
                ]);

            if ($productImage->isNew()) {
                $this->logger->info("Nuova immagine: " . $file->getFilename());
            }

            $destinationFile = FWUtils::addTrailingSlashes(config('catalog.images.dir')).$productImage->imported_filename;

            if ($this->copyFile($file, $destinationFile)) {
                // La ridimensiono/comprimo se necessario
                $this->resizeImage(new SplFileInfo($destinationFile));
            }

        }

    }

}
