<?php


namespace FourWords\AppManager\Catalog\Importer\Pim;

use DateTimeZone;
use Exception;
use FourWords\AppManager\Catalog\Importer\SyncableAbstractImporter;
use FourWords\AppManager\Catalog\Model\Category;
use FourWords\AppManager\Catalog\Model\CategoryProduct;
use FourWords\AppManager\Catalog\Model\Feature;
use FourWords\AppManager\Catalog\Model\FeatureProduct;
use FourWords\AppManager\Catalog\Model\Product;
use FourWords\AppManager\Catalog\Model\ProductsRelated;
use FourWords\AppManager\Catalog\Pim\PimApi;
use FourWords\AppManager\DataExtractor\DataExtractor;
use Illuminate\Support\Facades\DB;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Logger;

class PimCatalogImporter extends SyncableAbstractImporter
{

    /** @var PimCategoryImporter Creo le categorie dal Pim */
    protected $categoryImporter;

    /** @var PimFeatureImporter Creo le feature dal Pim */
    protected $featureImporter;

    protected $pimApi;

    /** @var Logger */
    public $logger;

    public function __construct(PimApi $pimApi, PimCategoryImporter $pimCategoryImporter, PimFeatureImporter $pimFeatureImporter, Product $productModel, DataExtractor $dataExtractor)
    {

        parent::__construct($productModel, $dataExtractor);

        $this->pimApi = $pimApi;
        $this->categoryImporter = $pimCategoryImporter;
        $this->featureImporter = $pimFeatureImporter;

        // Log
        $this->logger = new Logger('catalog-importer');
        $this->logger->setTimezone(new DateTimeZone('Europe/Rome'));

        if (!file_exists(config('catalog.log.dir'))) {
            mkdir(config('catalog.log.dir'),config('catalog.log.dirPermissions'), true);
        }

        // debug.log
        $LdebugHandler = new RotatingFileHandler(config('catalog.log.dir').'debug.log',14,Logger::DEBUG, true, config('catalog.log.filePermissions'));
        $debugDateFormat = "Y-m-d H:i:s";
        $debugOutput = "[%datetime%] %channel%.%level_name% : %message% \n";
        $debugFormatter = new LineFormatter($debugOutput, $debugDateFormat,true);
        $LdebugHandler->setFormatter($debugFormatter);
        $this->logger->pushHandler($LdebugHandler);
    }



    public function pushHandler($handler) {
        $this->logger->pushHandler($handler);
    }

    public function import() {

        // Creo gli alberi dalle categorie del PIM
        $this->categoryImporter->import();

        // Creo le feature dal PIM
        $this->featureImporter->import();

        $categoriesIdsByCode = Category::all()->groupBy('code')->map(function ($items) {
            return $items->pluck('id')->first();
        });

        $featuresIdsByCode = Feature::all()->groupBy('code')->map(function ($items) {
            return $items->pluck('id')->first();
        });

        // Recupero tutti i prodotti del pim
        $pimProducts = $this->pimApi->getProductsList(['visibility' => true]);
        // Raggruppo i prodotti del pim per id
        $productsById =  collect($pimProducts)->pluck('code', 'product_id');

        try {

            DB::beginTransaction();

            $this->model->setAllIsSynchronized(false);
            $this->truncate(CategoryProduct::class);
            $this->truncate(FeatureProduct::class);
            $this->truncate(ProductsRelated::class);

            $chunks = array_chunk($pimProducts,250);

            $categoriesByProductCode = [];
            $featuresByProductCode = [];

            foreach ($chunks as $chunk) {

                // recupero MassInfo dei prodotti
                $products = $this->pimApi->getMultipleProductMassInfo(["codes"=> array_column($chunk,'code')]);

                $productsArray = [];

                foreach ($products as $productMassiveInfos){

                    //divido le varie informazioni ricevute dal servizio
                    $productInfo = $productMassiveInfos['info'];
                    $productFeatures = $productMassiveInfos['featureProducts'] ?: $productMassiveInfos['featureValue'];
                    $productCategories = $productMassiveInfos['category'];
                    $productName = $productMassiveInfos['name']['it'];
                    $productDescription_short = $productMassiveInfos['description_short_clean']['it'] ?: $productMassiveInfos['description_short']['it'];
                    $productDescription = $productMassiveInfos['description_clean']['it'] ?: $productMassiveInfos['description']['it'];

                    $productsArray[] = [
                        'code' => $productInfo['code'],
                        'name' => $productName,
                        'short_description' => $productDescription_short,
                        'long_description' => $productDescription,
                    ];

                    if (!empty($productCategories)) $categoriesByProductCode[$productInfo['code']] = $productCategories;
                    if (!empty($productFeatures)) $featuresByProductCode[$productInfo['code']] = $productFeatures;
                }

                $this->sync($productsArray);

                $localProducts = Product::whereIn('code',collect($productsArray)->pluck('code')->toArray())->get();

                foreach ($productsArray as $product) {

                    $product = $localProducts->where('code',$product['code'])->first();

                    if (!$product) continue;

                    // Collego il prodotto all'ultima categoria
                    if ($productCategories = $categoriesByProductCode[$product->code]) {

                        foreach ($productCategories as $categoryTree) {
                            $latestCategory = last($categoryTree);
                            if($categoryId = $categoriesIdsByCode[$latestCategory['code']]) {
                                $product->categories()->attach($categoryId);
                            }
                        }
                    }

                    // Collego il prodotto alle feature TODO: Testare
                    if ($productFeatures = $featuresByProductCode[$product->code]) {

                        foreach ($productFeatures as $feature) {
                            if($featureId = $featuresIdsByCode[$feature['feature_id']]) {
                                $commaSeparatedValues = $feature['values']['it'];
                                $values = explode(",",$commaSeparatedValues);

                                foreach ($values as $v) {
                                    $product->features()->attach($featureId, ['value' => $v]);
                                }
                            }
                        }
                    }

                }

            }

            // Prendo la lista dei correlati di tutti i prodotti nel pim e la converto in codici, raggruppata per codice
            $relatedProductsList = $this->pimApi->getRelatedList();
            $relatedProductsByCodes = [];
            foreach ($relatedProductsList as $rel) {
                $productId = $rel['product_id'];
                $productRelatedId = $rel['product_related_id'];
                if ($productsById->has($productId) && $productsById->has($productRelatedId)) {
                    $relatedProductsByCodes[$productsById->get($productId)][] = $productsById->get($productRelatedId);
                }
            }

            // Prendo i prodotti in locale che devo collegare ai correlati
            $productsToRelate = Product::whereIn('code',array_keys($relatedProductsByCodes))->get();

            // Collego i prodotti correlati
            foreach ($productsToRelate as $product) {
                foreach ($relatedProductsByCodes[$product->code] as $relatedCode) {
                    $pRelated = Product::where('code',$relatedCode)->first();
                    if ($pRelated) $product->related()->attach($pRelated);
                }
            }

            $this->model->deleteNotSynchronized();

            DB::commit();

        } catch (Exception $e) {
            DB::rollBack();
            throw $e;
        }



    }

    public function getRecords() { }
}
