<?php


namespace FourWords\AppManager\Catalog\Importer\Pim;

use Exception;
use FourWords\AppManager\Catalog\Importer\AbstractFileImporter;
use FourWords\AppManager\Catalog\Model\ImageGallery;
use FourWords\AppManager\Catalog\Model\Product;
use FourWords\AppManager\Catalog\Model\ProductImage;
use FourWords\AppManager\Catalog\Model\SyncableModel;
use FourWords\AppManager\Catalog\Pim\PimApi;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Support\Facades\File;
use SplFileInfo;

class PimImageImporter extends AbstractFileImporter
{
    /** @var SyncableModel */
    protected $productImageModel;
    /** @var SyncableModel */
    protected $imageGalleryModel;

    public function __construct(ProductImage $productImageModel, ImageGallery $imageGalleryModel)
    {
        parent::__construct();
        $this->productImageModel = $productImageModel;
        $this->imageGalleryModel = $imageGalleryModel;
    }

    protected function getLoggerName(): string
    {
        return 'catalog-images';
    }

    protected function getDestinationPath(): string
    {
        return config('catalog.images.dir');
    }

    /**
     * @throws \Throwable
     */
    public function importFiles() {

        // Imposto le immagini esistenti nel db come "non sincronizzate",
        // nel momento in cui importerò le immagini, quelle che ho reimportato e quindi "sincronizzate" rimarranno e le "non sincronizzate" verranno eliminate
        $this->temporarilySetImagesAsNotSynchronized();

        // Importo le immagini come "sincronizzate"

        /** @var PimApi $pimApi */
        $pimApi = app()->make(PimApi::class);

        // Prendo tutte le colleciton di tipo immagine
        $collectionsMap = [];
        $types = collect($pimApi->getTypesList());
        $imageType = $types->firstWhere('slug', PimApi::TYPE_SLUG_IMAGES);
        if ($imageType) {
            $imageTypeId = $imageType['type_id'];
            $collections = collect($pimApi->getCollectionsList());
            $imageCollections = $collections->filter(function ($collection) use ($imageTypeId) {
                return $collection['type_id'] == $imageTypeId;
            });

            foreach($imageCollections as &$imageCollection) {
                // Se la galleria non esiste la creo, altrimenti imposto il flag synchronized a true
                /** @var SyncableModel $gallery */
                $gallery = $this->imageGalleryModel->sync(['name' => $imageCollection['name']['it']]);

                if ($gallery && $gallery->isNew()) {
                    $this->logger->info("Nuova galleria: ".$gallery->name);
                }

                $collectionsMap[$imageCollection['collection_id']] = $gallery->id;
            }

        }

        if (empty($collectionsMap)) throw new Exception("Impossibile recuperare le collection di immagini dal PIM");

        // Prendo tutte le gallery create localmente
//        $this->imageGalleryModel->all();

        // Prendo tutti i prodotti dal PIM
        $pimProducts = collect($pimApi->getProductsList(['visibility' => true]));

        // Filtro prendendo solamente i prodotti che ho localmente
        $productCodes = Product::whereIn('code', $pimProducts->pluck('code'))->pluck('code')->toArray();
        $pimProductsStoredLocally = array_filter($pimProducts->toArray(), function($item) use ($productCodes) {
            return in_array($item['code'], $productCodes);
        });

        foreach ($pimProductsStoredLocally as $pimProduct) {

            // Prendo la lista delle immagini di questo prodotto
            $productImages = $pimApi->getImagesList($pimProduct['product_id']);

            foreach ($productImages as $productImageToDownload) {

                // Se il record non esiste lo creo, altrimenti aggiorno i dati. Imposto il flag synchronized a true
                /** @var SyncableModel $productImage */
                $hashedFileName = $productImageToDownload['hash'].'.'.$productImageToDownload['extension'];
                $productImage = $this->productImageModel->sync(

                    ['product_code' => $pimProduct['code'], 'imported_filename' => $hashedFileName],
                    [
                        'filename' => $productImageToDownload['name']['it'],
                        'description' => $productImageToDownload['name']['it'],
                        'size' => 0.0,  // Non recuperata da PIM
                        'product_code' => $pimProduct['code'],
                        'position' => $productImageToDownload['position'],
                        'gallery' => $collectionsMap[$productImageToDownload['collection_id']]
                    ]);

                if ($productImage->isNew()) {
                    $this->logger->info("Nuova immagine: " . $productImageToDownload['name']['it']);
                }

                $destinationFile = FWUtils::addTrailingSlashes(config('catalog.images.dir')).$hashedFileName;

                if (!File::exists($destinationFile)) {
                    try {
                        $pimApi->downloadProductImageFile($productImageToDownload['image_product_id'], $destinationFile);
                        $this->logger->debug("Copiata immagine in: " . $destinationFile);
                    } catch (Exception $e) {
                        $this->logger->error('Impossibile scaricare l\'immagine da PIM con id '.$productImageToDownload['image_product_id'].' in: ' . $destinationFile);
                    }
                }

            }
        }


        $className = basename(str_replace('\\', '/', trim(get_class($this), '\\')));
        if ($errorMessage = $this->validateSynchronization()) {
            throw new Exception("$className :: validateSynchronization() ->" . $errorMessage);
        }

        // Elimino le immagini che non sono più "sincronizzate"
        $this->deleteImagesNotSynchronized();

        //TODO: Cancellare su filesystem le immagini non più sincronizzate con un task?
    }

    /**
     * Funzione per validare la sincronizzazione, la utilizzo ad esempio per evitare che le tabelle vengano totalmente (o quasi) svuotate se non trovo dei record per procedure lanciate da clienti sbarazzini o errori su servizi
     * @return false|string
     */
    protected function validateSynchronization()
    {
        if ($this->productImageModel->getNotSynchronizedPercentage() > 50) {
            return "La percentuale di file non sincronizzati supera il 50%";
        }

        return false;
    }

    protected function temporarilySetImagesAsNotSynchronized()
    {
        $this->imageGalleryModel->setAllIsSynchronized(false);
        $this->productImageModel->setAllIsSynchronized(false);
    }

    protected function deleteImagesNotSynchronized()
    {
        $this->imageGalleryModel->deleteNotSynchronized();
        $this->productImageModel->deleteNotSynchronized();
    }
     protected function importImages($importDir, $recursively = true, $maxLevel = 2, $level = 1,  $galleryName = '') {

        $files = File::files($importDir);
        $allFiles = $recursively ? array_merge(File::directories($importDir),$files) : $files;

        // Se la cartella specificata ha file o cartelle creo la collection, altrimenti non faccio nulla
        if (count($allFiles) == 0) {
            $this->logger->debug("Cartella ignorata perché non contiene file: ".basename($importDir));
            return;
        }

        // Se il nome galleria non è specificato, lo prendo dalla cartella che contiene il file
        if (empty($galleryName)) {
            $galleryName = basename($importDir);
        }
        $galleryName = ucfirst($galleryName);

        // Se la galleria non esiste la creo, altrimenti imposto il flag synchronized a true
        /** @var SyncableModel $gallery */
        $gallery = $this->imageGalleryModel->sync(['name' => $galleryName]);

        if ($gallery->isNew()) {
            $this->logger->info("Nuova galleria: ".$galleryName);
        }

        foreach ($allFiles as $file) {

            if (File::isDirectory($file)) {

                if ($maxLevel == -1 || $level < $maxLevel) $this->importImages($file, $recursively, $maxLevel, $level+1);
                continue;

            } else if ($file->isFile()) {

                $fileMatchRegularExpression = $this->getFileRegex();
                if ( !empty($fileMatchRegularExpression) && !preg_match($fileMatchRegularExpression,$file->getFilename()) ) {
                    continue;
                }

                $allowedExtensions = $this->getAllowedExtensions();
                if (!empty($allowedExtensions) && !in_array($file->getExtension(), $allowedExtensions)) {
                    continue;
                }
            }

            if (!$this->validateFileSizeInMegaBytes($file, config('catalog.images.maxSize'))) {
                continue;
            }

            try {
                $imageAttributes = $this->extractImageData($file);
                $imageAttributes->setCollection($gallery->id);
            } catch (Exception $e) {
                $this->logger->warning('Immagine non importata, impossibile recuperare i dettagli del file '.$file->getPathname());
                continue;
            }

            // Se il record non esiste lo creo, altrimenti aggiorno i dati. Imposto il flag synchronized a true
            /** @var SyncableModel $productImage */
            $productImage = $this->productImageModel->sync(
                ['size' => $imageAttributes->getSize(), 'filename' => $imageAttributes->getFilename(), 'gallery' => $gallery->id],
                [
                    'product_code' => $imageAttributes->getCode(),
                    'position' => $imageAttributes->getPosition(),
                    'gallery' => $gallery->id,
                    'imported_filename' => $this->generateImportedFilename($file)
                ]);

            if ($productImage->isNew()) {
                $this->logger->info("Nuova immagine: " . $file->getFilename());
            }

            $destinationFile = FWUtils::addTrailingSlashes(config('catalog.images.dir')).$productImage->imported_filename;

            if ($this->copyFile($file, $destinationFile)) {
                // La ridimensiono/comprimo se necessario
                $this->resizeImage(new SplFileInfo($destinationFile));
            }

        }

    }

    protected function getAllowedExtensions(): array
    {
        return ['jpg','jpeg','png','gif'];
    }
}
