<?php


namespace FourWords\AppManager\Catalog\Importer\Pim;

use Exception;
use FourWords\AppManager\Catalog\Importer\AbstractFileImporter;
use FourWords\AppManager\Catalog\Model\Product;
use FourWords\AppManager\Catalog\Model\ProductsHaveResources;
use FourWords\AppManager\Catalog\Model\Resource;
use FourWords\AppManager\Catalog\Model\ResourceCollection;
use FourWords\AppManager\Catalog\Model\SyncableModel;
use FourWords\AppManager\Catalog\Pim\PimApi;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\File;

class PimResourceImporter extends AbstractFileImporter
{
    /** @var SyncableModel */
    protected $resourceModel;
    /** @var SyncableModel */
    protected $resourceCollectionModel;

    protected $resourcesRootCollection;

    public function __construct(Resource $resourceModel, ResourceCollection $resourceCollectionModel)
    {
        parent::__construct();
        $this->resourceModel = $resourceModel;
        $this->resourceCollectionModel = $resourceCollectionModel;
    }


    protected function getLoggerName(): string{
        return 'catalog.resources';
    }

    protected function getDestinationPath(): string{
        return config('catalog.resources.dir');
    }

    /**
     * @throws \Throwable
     */
    public function importFiles() {

            // Imposto le risorse esistenti nel db come "non sincronizzate",
            // nel momento in cui importerò le risorse, quelle che ho reimportato e quindi "sincronizzate" rimarranno e le "non sincronizzate" verranno eliminate
            $this->temporarilySetResourcesAsNotSynchronized();
            $this->truncate(ProductsHaveResources::class);

            // Creo le collection di base per le risorse aziendali (Radice e Home)
            $this->createRootCollections();

            // Importo le risorse dei prodotti
            /** @var PimApi $pimApi */
            $pimApi = app()->make(PimApi::class);

            // Prendo tutte le colleciton di tipo documenti
            $collectionsMap = [];
            $types = collect($pimApi->getTypesList());
            $documentType = $types->firstWhere('slug', PimApi::TYPE_SLUG_DOCUMENTS);
            if ($documentType) {
                $documentTypeId = $documentType['type_id'];
                $collections = collect($pimApi->getCollectionsList());
                $documentsCollections = $collections->filter(function ($collection) use ($documentTypeId) {
                    return $collection['type_id'] == $documentTypeId;
                });

                foreach($documentsCollections as &$documentCollection) {
                    // Se la collection non esiste la creo, altrimenti imposto il flag synchronized a true
                    /** @var SyncableModel $resourceCollection */
                    $resourceCollection = $this->resourceCollectionModel->sync(['name' => $documentCollection['name']['it']]);

                    if ($resourceCollection && $resourceCollection->isNew()) {
                        $this->logger->info("Nuova collection di allegati: ".$resourceCollection->name);
                    }

                    $collectionsMap[$documentCollection['collection_id']] = $resourceCollection->id;
                }

            }

            if (empty($collectionsMap)) throw new Exception("Impossibile recuperare le collection di allegati prodotto dal PIM");

            // Prendo tutti i prodotti dal PIM
            $pimProducts = collect($pimApi->getProductsList(['visibility' => true]));

            $localProductsMap = Product::whereIn('code', $pimProducts->pluck('code'))->pluck('id','code')->toArray();

            $localProductsCodes = array_keys($localProductsMap);

            // TODO: Lasciare? Serve ad utilizzare solamente i prodotti pim che ho anche nel mysql. Ipotizzando di reperire con ProductImporter soltanto i prodotti di un determinato albero ad es...
            $pimProductsStoredLocally = array_filter($pimProducts->toArray(), function($item) use ($localProductsCodes) {
                return in_array($item['code'], $localProductsCodes);
            });

            foreach ($pimProductsStoredLocally as $pimProduct) {

                // Prendo la lista dei documenti di questo prodotto
                $productDocuments = $pimApi->getDocumentsList($pimProduct['product_id']);

                foreach ($productDocuments as $productDocumentToDownload) {

                    // Se il record non esiste lo creo, altrimenti aggiorno i dati. Imposto il flag synchronized a true
                    /** @var SyncableModel $productResource */
                    $hashedFileName = $productDocumentToDownload['hash'].'.'.$productDocumentToDownload['extension'];
                    $productResource = $this->resourceModel->sync(
                        ['imported_filename' => $hashedFileName],
                        [
                            'filename' => $productDocumentToDownload['name']['it'],
                            'type' => $productDocumentToDownload['extension'],
                            'description' => $productDocumentToDownload['name']['it'],
                            'size' => 0.0,  // Non recuperata da PIM
                            'position' => $productDocumentToDownload['position'],
                            'collection' => $collectionsMap[$productDocumentToDownload['collection_id']]
                        ]);

                    // Collego il record risorsa al prodotto
                    $productId = $localProductsMap[$pimProduct['code']];
                    if ($productId) {
                        $productResource->products()->attach($productId);
                    }

                    if ($productResource->isNew()) {
                        $this->logger->info("Nuova risorsa: " . $productDocumentToDownload['attachment_product_id']." del prodotto ".$pimProduct['code']);
                    }

                    $destinationFile = FWUtils::addTrailingSlashes(config('catalog.resources.dir')).$hashedFileName;

                    if (!File::exists($destinationFile)) {
                        try {
                            $pimApi->downloadProductDocumentFile($productDocumentToDownload['attachment_product_id'], $destinationFile);
                            $this->logger->debug("Copiata risorsa in: " . $destinationFile);
                        } catch (Exception $e) {
                            $this->logger->error('Impossibile scaricare risorsa da PIM con id '.$productDocumentToDownload['attachment_product_id'].' in: ' . $destinationFile);
                        }
                    }

                }
            }


            $className = basename(str_replace('\\', '/', trim(get_class($this), '\\')));
            if ($errorMessage = $this->validateSynchronization()) {
                throw new Exception("$className :: validateSynchronization() ->" . $errorMessage);
            }

            // Elimino le risorse che non sono più "sincronizzate"
            $this->deleteResourcesNotSynchronized();
    }

    protected function validateSynchronization()
    {
        if ($this->resourceModel->getNotSynchronizedPercentage() > 50) {
            return "La percentuale di file non sincronizzati supera il 50%";
        }

        return false;
    }

    protected function truncate($model) {

        if (is_string($model))
            $model = new $model();

        $model->getConnection()->statement('SET FOREIGN_KEY_CHECKS = 0;');
        DB::table($model->getTable())->truncate();
        $model->getConnection()->statement('SET FOREIGN_KEY_CHECKS = 1;');
    }

    protected function temporarilySetResourcesAsNotSynchronized()
    {
        $this->resourceCollectionModel->setAllIsSynchronized(false);
        $this->resourceModel->setAllIsSynchronized(false);
    }

    protected function deleteResourcesNotSynchronized()
    {
        $this->resourceCollectionModel->deleteNotSynchronized();
        $this->resourceModel->deleteNotSynchronized();
    }

    protected function getAllowedExtensions(): array
    {
        return ['jpg','jpeg','png','gif','pdf','doc','docx'];
    }

    protected function createRootCollections()
    {
        $radice = $this->resourceCollectionModel->sync(['name' => 'Radice']);
        $this->resourcesRootCollection = $this->resourceCollectionModel->sync(['name' => 'Home', 'parent_id' => $radice->id]);
    }

}
