<?php


namespace FourWords\AppManager\Catalog\Importer;

use FourWords\AppManager\Catalog\Model\ResourceCollection;
use FourWords\AppManager\Catalog\Model\Resource;
use FourWords\AppManager\Catalog\Model\SyncableModel;
use FourWords\AppManager\Catalog\Resources\ResourceAttributes;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Support\Facades\File;
use SplFileInfo;

class ResourceImporter extends AbstractFileImporter
{
    /** @var SyncableModel */
    protected $resourceModel;
    /** @var SyncableModel */
    protected $resourceCollectionModel;

    protected $resourcesRootCollection;

    public function __construct(Resource $resourceModel, ResourceCollection $resourceCollectionModel)
    {
        parent::__construct();
        $this->resourceModel = $resourceModel;
        $this->resourceCollectionModel = $resourceCollectionModel;
    }


    protected function getLoggerName(): string{
        return 'catalog.resources';
    }

    protected function getDestinationPath(): string{
        return config('catalog.resources.dir');
    }

    /**
     * @throws \Throwable
     */
    public function importFiles() {

            // Imposto le risorse esistenti nel db come "non sincronizzate",
            // nel momento in cui importerò le risorse, quelle che ho reimportato e quindi "sincronizzate" rimarranno e le "non sincronizzate" verranno eliminate
            $this->temporarilySetResourcesAsNotSynchronized();

            // Creo le collection di base (Radice e Home)
            $this->createRootCollections();

            // Importo le risorse con parent "Home" e nome "Risorse aziendali"
            $this->importResources(config('catalog.resources.importDir'),true, -1, 1, $this->resourcesRootCollection->parent_id,$this->resourcesRootCollection->name);

            // Elimino le risorse che non sono più "sincronizzate"
            $this->deleteResourcesNotSynchronized();
    }

    protected function temporarilySetResourcesAsNotSynchronized()
    {
        $this->resourceCollectionModel->setAllIsSynchronized(false);
        $this->resourceModel->setAllIsSynchronized(false);
    }

    protected function deleteResourcesNotSynchronized()
    {
        $this->resourceCollectionModel->deleteNotSynchronized();
        $this->resourceModel->deleteNotSynchronized();
    }

    /**
     * @param SplFileInfo $file
     * @return ResourceAttributes
     */
    protected function extractResourceData(SplFileInfo $file): ResourceAttributes
    {
        return new ResourceAttributes($file->getFilename(), $file->getSize(), 1, strtolower($file->getExtension()));
    }

    protected function getAllowedExtensions(): array
    {
        return ['jpg','jpeg','png','gif','pdf','doc','docx'];
    }

    protected function createRootCollections()
    {
        $radice = $this->resourceCollectionModel->sync(['name' => 'Radice']);
        $this->resourcesRootCollection = $this->resourceCollectionModel->sync(['name' => 'Home', 'parent_id' => $radice->id]);
    }

    protected function importResources($importDir, $recursively = true, $maxLevel = -1, $level = 1, $parentId = null, $collectionName = '') {

        $files = File::files($importDir);
        $allFiles = $recursively ? array_merge(File::directories($importDir),$files) : $files;

        // Se la cartella specificata ha file o cartelle creo la collection, altrimenti non faccio nulla
        if (count($allFiles) == 0) {
            $this->logger->debug("Cartella ignorata perché non contiene file: ".basename($importDir));
            return;
        }

        // Se il nome della collection non è specificato, lo prendo dalla cartella che contiene il file
        if (empty($collectionName)) {
            $collectionName = basename($importDir);
        }
        $collectionName = ucfirst($collectionName);

        // Se la collection non esiste la creo, altrimenti imposto il flag synchronized a true
        /** @var SyncableModel $collection */
        $collection = $this->resourceCollectionModel->sync(['name' => $collectionName, 'parent_id' => $parentId]);

        if ($collection->isNew()) {
            $this->logger->info("Nuova collection: ".$collectionName);
        }

        $positionCounter = 0;

        foreach ($allFiles as $file) {

            if (File::isDirectory($file)) {

                $parentId = $collection->id;
                if ($maxLevel == -1 || $level < $maxLevel) $this->importResources($file, $recursively, $maxLevel, $level+1, $parentId);
                continue;

            } else if ($file->isFile()) {

                $fileMatchRegularExpression = $this->getFileRegex();
                if ( !empty($fileMatchRegularExpression) && !preg_match($fileMatchRegularExpression,$file->getFilename()) ) {
                    continue;
                }

                $allowedExtensions = $this->getAllowedExtensions();
                if (!empty($allowedExtensions) && !in_array($file->getExtension(), $allowedExtensions)) {
                    continue;
                }
            }

            $positionCounter += 1;

            if (!$this->validateFileSizeInMegaBytes($file, config('catalog.resources.maxSize'))) {
                continue;
            }

            $resourceAttributes = $this->extractResourceData($file);
            $resourceAttributes->setPosition($positionCounter);

            // Se il record non esiste lo creo, altrimenti aggiorno i dati. Imposto il flag synchronized a true
            /** @var SyncableModel $resource */
            $resource = $this->resourceModel->sync(
                ['size' => $resourceAttributes->getSize(), 'filename' => $resourceAttributes->getFilename(), 'collection' => $collection->id],
                [
                    'position' => $resourceAttributes->getPosition(),
                    'type' => $resourceAttributes->getType(),
                    'imported_filename' => $this->generateImportedFilename($file)
                ]);

            if ($resource->isNew()) {
                $this->logger->info("Nuova risorsa: " . $file->getFilename());
            }

            // Copio la risorsa nel filesystem se non esiste già
            $destinationFile = FWUtils::addTrailingSlashes(config('catalog.resources.dir')) . $resource->imported_filename;
            $this->copyFile($file, $destinationFile);

        }

    }

}
