<?php

namespace FourWords\AppManager\Catalog\Importer;

use Exception;
use FourWords\AppManager\Catalog\Model\SyncableModel;
use FourWords\AppManager\DataExtractor\DataExtractor;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Support\Facades\DB;

abstract class SyncableAbstractImporter extends AbstractImporter
{

    /** @var SyncableModel */
    protected $model;

    public function __construct(SyncableModel $model, DataExtractor $dataExtractor)
    {
        parent::__construct($model, $dataExtractor);
    }

    public function import() {

        try {

            $records = $this->getRecords();
            $this->validateRecords($records);

            DB::beginTransaction();

            $this->model->setAllIsSynchronized(false);
            $this->sync($records);
            $this->model->deleteNotSynchronized();

            DB::commit();

        } catch (Exception $e) {
            DB::rollBack();
            throw $e;
        }
    }

    /**
     * Funzione per validare i record, la utilizzo ad esempio per evitare che le tabelle vengano totalmente (o quasi) svuotate se non trovo dei record per procedure lanciate da clienti sbarazzini o errori su servizi
     * @param $records
     * @return void
     * @throws Exception
     */
    protected function validateRecords($records)
    {
        $className = basename(str_replace('\\', '/', trim(get_class($this), '\\')));

        $localCount = $this->model->count();
        $recordsCount = count($records);

        if ($recordsCount < $localCount) {

            // Percentuale di record recuperati sul totale che avevo
            $percentage = ($recordsCount / $localCount) * 100;

            if ($percentage < 50) {
                throw new Exception("$className :: I record recuperati non superano il 50% dei record che ho in locale");
            }
        }
    }

    protected function normalizeKey(&$key)
    {
        return str_replace('\\', '', $key);
    }

    protected function sync($rows) {

        if (empty($this->model->distinctKey)) {
            throw new Exception("Specificare chiavi per SyncableModel: ".get_class($this->model));
        }

        $modelsByDistinctKey = $this->model->all()->keyBy($this->model->distinctKey);

        if ($modelsByDistinctKey->count() == 0) {
            $this->insert($rows);
            return;
        }

        $newRows = [];
        $idsToSetSynchronized = [];

        foreach ($rows as $row) {
            if ($model = $modelsByDistinctKey[$this->normalizeKey($row[$this->model->distinctKey])]) {


                $record = FWUtils::removeNullsAndEmptyString($row);
                $model->fill($record);

                if (!empty($model->getDirty())) {
                    // Se ci sono cambiamenti eseguo l'update della riga con i nuovi dati
                    $model->{$model->isSyncedProperty} = true;
                    $model->save();
                } else {
                    $idsToSetSynchronized[] = $model->{$this->model->getKeyName()};
                }
            } else {
                $newRows[] = $row;
            }
        }

        // Imposto come synchronized = true i model che esistevano ma che non ho dovuto aggiornare
        $this->model->whereIn($this->model->getKeyName(), $idsToSetSynchronized)
            ->update([$this->model->isSyncedProperty => true]);

        // Creo i nuovi record
        if (count($newRows) > 0) {
            $this->insert($newRows);
        }

    }

    protected function onInsertRecord(&$record)
    {
        $record->{$this->model->isSyncedProperty} = true;
    }

}