<?php

namespace FourWords\AppManager\Catalog\Pim;

use Exception;
use FourWords\AppManager\Director\FWUtils;
use Illuminate\Support\Facades\Http;

class PimApi
{

    protected string $apiUrl;
    protected $version;
    protected $user;
    protected $password;

    protected $token;

    protected array $headers = [];

    protected $connectionAttempts = 0;
    protected $maxConnectionAttempts = 3;

    //Slug delle tipologie
    const TYPE_SLUG_IMAGES = 'images';
    const TYPE_SLUG_DOCUMENTS = 'documents';
    const TYPE_SLUG_VIDEOS = 'videos';
    const TYPE_SLUG_FEATURES = 'features';
    const TYPE_SLUG_PRODUCTS = 'products';

    public function __construct()
    {
        $this->setApiUrl(config('catalog.pim.apiUrl'));
        $this->setUser(config('catalog.pim.user'));
        $this->setPassword(config('catalog.pim.password'));
    }

    /**
     * @return mixed
     */
    public function getCompleteApiUrl()
    {
        return FWUtils::addTrailingSlashes($this->apiUrl.(!empty($this->getVersion()) ? '/'.$this->getVersion() : ''));
    }

    /**
     * @param mixed $baseUrl
     */
    public function setApiUrl($apiUrl): void
    {
        $this->apiUrl = $apiUrl;
    }

    /**
     * @return mixed
     */
    public function getVersion()
    {
        return $this->version;
    }

    /**
     * @param mixed $version
     */
    public function setVersion($version): void
    {
        $this->version = $version;
    }



    /**
     * @return mixed
     */
    public function getUser()
    {
        return $this->user;
    }

    /**
     * @param mixed $user
     */
    public function setUser($user): void
    {
        $this->user = $user;
    }

    /**
     * @return mixed
     */
    public function getPassword()
    {
        return $this->password;
    }

    /**
     * @param mixed $password
     */
    public function setPassword($password): void
    {
        $this->password = $password;
    }

    /**
     * @return mixed
     */
    public function getToken()
    {
        return $this->token;
    }

    /**
     * @param mixed $token
     */
    public function setToken($token): void
    {
        $this->token = $token;
    }

    /**
     * @param $key
     * @param $value
     */
    protected function addHeader($key,$value) {
        $this->headers[$key] = $value;
    }

    /**
     * @param array $headers
     */
    protected function addHeaders(array $headers = array()) {
        foreach ($headers as $key => $value) {
            $this->addHeader($key,$value);
        }
    }

    protected function getHeaders() {

        $headers = $this->headers;

        if(!$this->token) {
            $this->authenticate();
        }

        //aggiungo la chiave nella lista
        $headers['Authorization'] = 'Bearer '.$this->token;
        $headers['User-Agent'] = 'Mozilla/5.0';
        return $headers;
    }

    protected function authenticate()
    {
        $url = $this->getCompleteApiUrl() . 'user/login';

        try {
            $response = Http::withHeaders([
                'User-Agent' => 'Mozilla/5.0',
            ])->post($url, [
                'username' => $this->getUser(),
                'password' => $this->getPassword(),
            ])->throw();

            $token = $response->json('token.access_token');

            if (!$token) {
                throw new Exception(__METHOD__.": Impossibile recuperare il token dal PIM. URL: $url");
            }

            $this->token = $token;
        } catch (Exception $exception) {
            throw new Exception(__METHOD__.": Impossibile effettuare richiesta al PIM. URL: $url\n".$exception->getMessage());
        }
    }

    protected function get($request, $isJson = true)
    {
        $request = ltrim($request, '/');
        $url = $this->getCompleteApiUrl() .'pim/'. $request;

        try {
            $response = Http::withHeaders($this->getHeaders())
                ->get($url);

            // Handle redirection or unauthorized response (e.g., expired token)
            if ($response->status() == 302 || $response->status() == 401) {
                // Check if there are remaining attempts
                if (++$this->connectionAttempts < $this->maxConnectionAttempts) {
                    // Riprovo la richiesta
                    $this->token = null;
                    return $this->get($request);
                } else {
                    throw new Exception(__METHOD__.":Impossibile eseguire richiesta PIM. Troppi tentativi raggiunti (".$this->maxConnectionAttempts.") URL: $url");
                }
            }

            // Check for errors
            $response->throw();

            // Reset connection attempts
            $this->connectionAttempts = 0;

            // Clear headers used for the request
            $this->headers = [];

            return $isJson ? $response->json() : $response->body();
        } catch (Exception $exception) {
            throw new Exception(__METHOD__.": Impossibile eseguire richiesta PIM. Status: URL: $url\n".$exception->getMessage());
        }
    }

    public function getProductsList($filters = array()) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/products');
    }

    public function getFeaturesList() {
        return $this->get('/features');
    }

    public function getProducts($id) {

        if(!$id) throw new Exception(sprintf("%s: parametro id richiesto",__METHOD__));

        return $this->get(sprintf('/product/%s',$id));
    }

    public function getProductFeatures($id) {

        if(!$id) throw new Exception(sprintf("%s: parametro id richiesto",__METHOD__));

        return $this->get(sprintf('/product/%s/feature',$id));
    }

    public function getProductCategories($id) {

        if(!$id) throw new Exception(sprintf("%s: parametro id richiesto",__METHOD__));

        return $this->get(sprintf('/product/%s/category',$id));
    }

    public function getCategoriesList($filters = array()) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/categories');
    }

    public function getRelatedList() {
        return $this->get('/related');
    }

    public function getImagesList($product, $filters = null) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/image/product/'.($product ?? ""));
    }

    public function getCategoriesImagesList($filters) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/image/category');
    }

//    public function getDownloadImageUrl($idImage) {
//        return $this->getCompleteApiUrl().'pim/download/image/'.$idImage;
//    }

    public function getTypesList() {
        return $this->get('/types');
    }

    public function getCollectionsList() {
        return $this->get('/collections');
    }

    public function getDocumentsList($product, $filters = null) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/document/product/'.($product ?? ""));
    }

    public function getCategoriesDocumentsList($filters) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/document/category');
    }

//    public function getDownloadDocumentUrl($idDocument) {
//        return $this->getCompleteApiUrl().'pim/download/'.$idDocument;
//    }

    public function getVideosList($filters) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/video/product');
    }

    public function getCategoriesVideosList($filters) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get('/video/category');
    }

//    public function getDownloadVideoUrl($idVideo) {
//        return $this->getCompleteApiUrl().'pim/download/'.$idVideo;
//    }

    /**
     * @param $url
     * @param $destination
     * @return true
     * @throws Exception
     */
    protected function downloadFile($url, $destination) {
        $wasSaved = file_put_contents($destination, $this->get($url, false));

        if (!$wasSaved) {
            throw new Exception("Impossibile salvare il file dal PIM:".$url."\nNella destinazione: ".$destination);
        }
        return true;
    }

    public function downloadProductImageFile($productImageId, $destination) {
        $url = "/download/image/product/".($productImageId ?? "");
        $wasSaved = file_put_contents($destination, $this->get($url, false));

        if (!$wasSaved) {
            throw new Exception("Impossibile salvare immagine dal PIM:".$url."\nNella destinazione: ".$destination);
        }
        return true;
    }

    public function downloadProductDocumentFile($productAttachmentId, $destination) {
        $url = "/download/document/product/".($productAttachmentId ?? "");
        $wasSaved = file_put_contents($destination, $this->get($url, false));

        if (!$wasSaved) {
            throw new Exception("Impossibile salvare attachment dal PIM:".$url."\nNella destinazione: ".$destination);
        }
        return true;
    }

    public function getProductMassInfo($id) {

        if(!$id) throw new Exception(sprintf("%s: parametro id richiesto",__METHOD__));

        return $this->get(sprintf('/product/%s/massive',$id))[0];
    }

    public function getMultipleProductMassInfo($filters) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get(sprintf('/product/massive'));
    }

    public function getCategory($id) {

        if(!$id) throw new Exception(sprintf("%s: parametro id richiesto",__METHOD__));

        return $this->get(sprintf('/category/%s',$id))[0];
    }

//    public function getDownloadCategoryUrl($idCategory) {
//        return $this->getCompleteApiUrl().'pim/download/'.$idCategory;
//    }

    public function getNewsList() {
        return $this->get('/news');
    }

    public function getUsers($filters) {

        if($filters) {
            $this->addHeader('filters', json_encode($filters));
        }

        return $this->get(sprintf('/users'));
    }


}
