<?php
/**
 * Copyright (c) 2020/3
 * Created by Federico Papetti
 */

namespace FourWords\AppManager\Director;


use DirectoryIterator;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Str;

class FWUtils
{

    private function __construct() {}

    public static function getFreeSpace($path = ".") {

        $freeSpace = disk_free_space($path);

        $string = self::humanFilesize($freeSpace);
        $totalSpace = disk_total_space($path);
        $string .= " / " . self::humanFilesize($totalSpace);

        return $string;
    }

    public static function humanFilesize($bytes, $decimals = 2) {
        $si_prefix = array( 'B', 'KB', 'MB', 'GB', 'TB', 'EB', 'ZB', 'YB' );
        $base = 1024;

        $class = min((int)log($bytes , $base) , count($si_prefix) - 1);
        return  sprintf("%1.{$decimals}f" , $bytes / pow($base,$class)) . ' ' . $si_prefix[$class];
    }

    public static function humanSeconds($seconds) {
        $hours = floor($seconds / 3600);
        $minutes = floor($seconds % 3600 / 60);
        $seconds = $seconds % 60;

        if($hours>0)
        {
            return sprintf(" %d h %02d , m %02d , s", $hours, $minutes, $seconds);
        }
        else if($minutes>0)
        {
            return sprintf("%02d m , %02d s", $minutes, $seconds);
        }
        else
        {
            return sprintf("%02d s",  $seconds);
        }
    }

    /**
     * Rimuove in modo ricorsivo tutti i file e le cartelle contenute nel percorso passato
     * @param string $dir
     */
    public static function recursiveDeleteDir($dir)
    {
        if (file_exists($dir)) {
            foreach (new DirectoryIterator($dir) as $fileInfo) {
                if (!$fileInfo->isDot()) {
                    if ($fileInfo->isDir()) {
                        static::recursiveDeleteDir($fileInfo->getPathname());
                    } else {
                        unlink($fileInfo->getPathname());
                    }
                }
            }
            rmdir($dir);
        }
    }

    /**
     * Sostituisce i bindings nella query, esempio:
     * SELECT * FROM table WHERE field = ? ---> SELECT * FROM table WHERE field = 'abcde'
     * @param $query
     * @param $bindings
     * @return string
     */
    public static function replaceBindingsInQuery($query, $bindings) {
        if (!empty($bindings)) {
            $query =  vsprintf(str_replace('?', '%s', $query), collect($bindings)->map(function ($binding) {
                return is_numeric($binding) ? $binding : "'{$binding}'";
            })->toArray());
        }
        return $query;
    }

    /**
     * Aggiunge ' o " all'inizio e la fine di una stringa o di tutti i valori di un array
     * @param $mixed
     * @param bool $single Se quote singolo (') o doppio (")
     * @return string|string[]
     */
    public static function inQuotes($mixed, bool $single = true, $escape = true, $emptyStringToNull = true, $nullToString = false) {

        $quote = $single ? "'" : '"';

        // Se non è un array
        if (!is_array($mixed)) {
            if ($nullToString && $mixed === null) {
                $mixed = self::nullToEmptyString($mixed);   // Ritorno stringa vuota da null
            } else if ($emptyStringToNull && $mixed === '') {
                return self::emptyStringToNull($mixed);     // Ritorno null da stringa vuota
            } else if ($mixed === null) {
                return 'null';                                // Ritorno null se null
            }

            // Aggiungo i quotes
            $result = $escape ? self::singleQuotesSqliteEscaped($mixed) : $mixed;
            return $quote.$result.$quote;
        }

        // Se è un array
        $mixed = (array) $mixed;
        return array_map(function($value) use ($quote, $escape, $nullToString, $emptyStringToNull) {

            if ($nullToString && $value === null) {
                $value = self::nullToEmptyString($value);   // Ritorno stringa vuota da null
            } else if ($emptyStringToNull && $value === '') {
                return self::emptyStringToNull($value);     // Ritorno null da stringa vuota
            } else if ($value === null) {
                return 'null';                                // Ritorno null se null
            }


            $result = $escape ? self::singleQuotesSqliteEscaped($value) : $value;
            return $quote.$result.$quote;
        }, $mixed);

    }

    /**
     * Converto stringa vuota in NULL.
     * Se è un array lo faccio per tutti gli elementi
     *
     * Mi serve per convertire in NULL le stringhe vuote che arrivano dall'AS400, poiché
     * anche se ho definito una colonna come Nullable, avrebbe valore "" invece di NULL
     * @param string|array
     * @return string|null|array
     */
    public static function emptyStringToNull($mixed) {

        if (!is_array($mixed)) {
            return $mixed === "" ? 'null' : $mixed;
        }

        $mixed = (array) $mixed;
        return array_map(function($value) {
            return $value === "" ? 'null' : $value;
        }, $mixed);
    }

    public static function removeNulls($array) {

        $array = (array) $array;
        return array_filter($array, function($value) {
            return !($value === "null");
        });
    }

    public static function removeNullsAndEmptyString($array) {

        $array = (array) $array;
        return array_filter($array, function($value) {
            return !(($value === null || $value === "null" || $value === ""));
        });
    }

    public static function nullToEmptyString($mixed) {

        if (!is_array($mixed)) {
            return $mixed === null ? '' : $mixed;
        }

        $mixed = (array) $mixed;
        return array_map(function($value) {
            return $value === null ? '' : $value;
        }, $mixed);
    }

    public static function singleQuotesSqliteEscaped($mixed) {

        if (!is_array($mixed) && $mixed === null) {
            return null;
        }

        if (!is_array($mixed)) {
            return str_replace("'","''",$mixed);
        }

        $mixed = (array) $mixed;
        return array_map(function($value) {
            return str_replace("'","''",$value);
        }, $mixed);
    }

    public static function asDateToTimestamp($asDateYmd, $hour = 0,$minute = 0, $second = 0) {
        if((int)$asDateYmd != 0)
            return gmmktime($hour,$minute,$second,substr($asDateYmd, 4,2),substr($asDateYmd, 6,2),substr($asDateYmd, 0,4));
        return 0;
    }

    /**
     * Controllo che l'oggetto passato abbia il trait, passare NomeTrait::class
     * @param $object
     * @param string $traitName
     * @return bool
     */
    public static function hasTrait( $object, $traitName )
    {
        $ret = class_uses($object);
        if( is_array( $ret ) )
        {
            $ret = array_search( $traitName, $ret ) !== false;
            return $ret;
        }
        return true;
    }

//    /**
//     * Controllo che la classe abbia sovrascritto il metodo della classe parent
//     * @param $subClassName
//     * @param $methodName
//     * @return bool
//     */
//    public static function hasOverriddenMethod($subClassName, $methodName)
//    {
//        try {
//            $foo = new ReflectionMethod($subClassName, $methodName);
//            $declaringClass = $foo->getDeclaringClass()->getName();
//            $proto = $foo->getPrototype();
//
//        } catch (\ReflectionException $e) {
//            return false;
//        }
//
//        return ($proto && $proto->getDeclaringClass()->getName() !== $declaringClass);
//    }

    /**
     * Rimuovo lo / finale in una stringa
     * @param $string
     * @return string
     */
    public static function removeTrailingSlashes($string)
    {
        return rtrim($string, '/');
    }

    /**
     * Aggiungo lo / finale in una stringa
     * @param $string
     * @return string
     */
    public static function addTrailingSlashes($string)
    {
        return rtrim($string, '/').'/';
    }

    /**
     * Aggiungo o sostituisco uno script in un file dato un determinato placeholder
     *
     * es. nel file
     *
     * -- test123 --
     * testscript
     * -- test123 --
     * abc
     *
     * addOrReplaceScript('file.js','-- test123 --',$script,'abc')
     * Aggiunge/sostituisce lo script prima di 'abc'
     *
     * addOrReplaceScript('file.js','-- test123 --',$script)
     * Aggiunge/sostituisce lo script a fine file
     *
     */
    public static function addOrReplaceScript($filePath, $placeholder, $script, $beforeText = null) {

        if ($contents = file($filePath)) {

$script = "\n".$placeholder."\n".$script."\n".$placeholder."\n";
            $scriptRows = substr_count($script,"\n");

            $scriptStartIndex = -1;
            $scriptAlreadyInFile = false;

            for($i = 0; $i < count($contents); $i++) {
                if (Str::contains($contents[$i], $placeholder)) {
                    $scriptStartIndex = $i;
                    $scriptAlreadyInFile = true;
                    break;
                } else if ($beforeText != null && Str::contains($contents[$i], $beforeText)) {
                    $scriptStartIndex = $i;
                    break;
                }
            }

            // Se non ho trovato lo script o il testo che cercavo esco
            if ($scriptStartIndex == -1) {
                return false;
            }

            array_splice($contents, $i-1, $scriptAlreadyInFile ? $scriptRows : 0, $script);
            File::put($filePath, $contents);
            return true;

        }

        return false;

    }

    /**
     *
     * Rimuovo una riga da un file
     *
     * @param $filePath
     * @param $placeholder
     * @param $script
     * @param null $beforeText
     * @return bool
     */
    public static function removeRow($filePath, $row, $exactMatch = true) {

        if ($contents = file($filePath)) {

            for($i = 0; $i < count($contents); $i++) {
                if ( ($exactMatch && $contents[$i] == $row) || (!$exactMatch && Str::contains($contents[$i],$row)) ) {
                    array_splice($contents, $i, 1);
                    File::put($filePath, $contents);
                    return true;
                }
            }

            return false;

        }

        return false;

    }
}
