<?php

namespace FourWords\AppManager\Ioc\Bonjour;

use Carbon\Carbon;
use DateTimeZone;
use FourWords\AppManager\Ioc\Contracts\Bonjour;
use FourWords\AppManager\Ioc\Models\BonjourHistory;
use FourWords\AppManager\PushNotifications\Model\PushNotificationSent;
use FourWords\AppManager\PushNotifications\Model\PushNotificationSentUser;
use Illuminate\Auth\Authenticatable;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\Relations\Pivot;
use Illuminate\Support\Facades\DB;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Logger;

abstract class AbstractBonjour implements Bonjour
{

    /** @var Logger  */
    protected $logger;

    /** @var array[] */
    protected $response;

    /** @var Authenticatable */
    protected $user;
    /** @var Model */
    protected $device;

    /** @var Pivot */
    protected $userDevicePivot;

    /** @var string */
    protected $token;

    protected $authenticatingWithToken = false;

    public function __construct()
    {

        // Log
        $this->logger = new Logger('bonjour');
        $this->logger->setTimezone(new DateTimeZone('Europe/Rome'));

        if (!file_exists(config('ioc.log.dir'))) {
            mkdir(config('ioc.log.dir'),config('ioc.log.dirPermissions'), true);
        }

        // debug.log
        $LdebugHandler = new RotatingFileHandler(config('ioc.log.dir').'bonjour/debug.log',14,Logger::DEBUG, true, config('ioc.log.filePermissions'));
        $debugDateFormat = "Y-m-d H:i:s";
        $debugOutput = "[%datetime%] %channel%.%level_name% : %message% \n";
        $debugFormatter = new LineFormatter($debugOutput, $debugDateFormat,true);
        $LdebugHandler->setFormatter($debugFormatter);
        $this->logger->pushHandler($LdebugHandler);


        $this->response = [
            'data' => [
                'now' => Carbon::now()->timestamp
            ],
            'status' => 'OK'
        ];
    }

    public function shouldAuthenticateRemotely()
    {
        return false;
    }

    public function authenticateRemotely()
    {
        return false;
    }

    public function getToken() {
        return request()->bearerToken();
    }

    public function bonjour() {

        try {

            // Controllo che ci siano le licenze
            if (!$this->hasLicenses()) {
                $this->report('Licenza d\'uso scaduta', Logger::CRITICAL);
                return $this->sendResponse('LICENSE_EXPIRED', 'Licenza d\'uso scaduta');
            }

            $this->user = $this->authenticate();

            if (!$this->user) {
                return $this->sendResponse('USER_NO_AUTH', 'Username e/o password errati');
            }

            if (!$this->isDemo() && !$this->isUserEnabled() ) {
                return $this->sendResponse('DISABLED_USER', 'Utente non abilitato');
            }

            if ($this->shouldAuthenticateRemotely()) {
                $this->response = $this->authenticateRemotely();
            }

            if ($this->response['status'] != 'OK') {
                return $this->sendResponse($this->response['status'], $this->response['data']['error'] ?? '');
            }

            $this->device = $this->getDevice();

            // Se il device è già registrato nel database
            if ($this->device) {

                if ($this->shouldWipe()) {
                    $this->report('L\'utente ' . request('user') . ' deve eseguire il wipe su dispositivo '.$this->device->name);
                    return $this->sendResponse('WIPE', 'Dispositivo non abilitato');
                }

                if (!$this->isDeviceEnabled()) {
                    $this->report('L\'utente ' . request('user') . ' ha il device '.$this->device->name.' disabilitato');
                    return $this->sendResponse('DEV_NO_AUTH', 'Dispositivo non abilitato');
                }

                //TODO: Clean?
//                $this->response['data']['clean'] = $this->shouldClean();


            } else {
                $this->createDevice();
            }

            $this->userDevicePivot = $this->getCurrentUserDevice();

            if (!$this->userDevicePivot) {
                $this->attachUserToDevice();
                $this->userDevicePivot = $this->getCurrentUserDevice();
            }

            //se lo stato della coppia utente/dispositivo è disabilitata mando json con status disabled_device_user
            if (!$this->isUserDeviceEnabled()) {
                return $this->sendResponse('USER_DEV_NO_AUTH', 'Utente non abilitato all\'uso del dispositivo');
            }

            if ($this->shouldWaitApproval()) {
                $this->report("L'utente con codice " . $this->user->code . " - " . $this->user->name . " è in attesa del permesso per l'utilizzo del dispositivo $this->device->name");
                $this->setUserDeviceStatus('WAIT');
            }

            //se lo stato della coppia utente/dispositivo è in WAIT mando json con status wait a meno che non sia in whitelist
            if ($this->userDevicePivot->status === 'WAIT') {
                return $this->sendResponse('WAIT_AUTH', 'In attesa di abilitazione da parte dell\'amministratore');
            }

            $this->token = $this->authenticatingWithToken ? $this->getToken() : $this->createToken();

            $this->report("L'utente con codice ". $this->user->code . " - " . $this->user->name . " ha effettuato l'accesso con il dispositivo " . $this->device->name . " e versione ".request('app_version', ''), Logger::INFO);

            $this->updateUserDeviceLastAccess();
            $this->updateDevice();
            $this->addBonjourToHistory();

            $this->updatePushNotificationsStats();

            $payload = ['data' => $this->getUserData(), 'token' => $this->token];

            return $this->sendResponse('OK', '', $payload);

        } catch (\Exception | \Throwable $e ) {
            $this->onException($e);
            throw($e);
        }

    }

    public function onException($e)
    {
        if (isset($this->logger)) {
            $this->logger->error('========================');
            $this->logger->error('Bonjour Fallito, $_REQUEST:');
            $this->logger->error(json_encode($_REQUEST));
            $this->logger->error('### Exception:');
            $this->logger->error($e->getMessage());
            $this->logger->error('========================');
            telegramError($e->getMessage());
        }
    }

    public function report($message, $gravity = Logger::WARNING) {

        $this->logger->log($gravity, $message);

        if ($gravity >= Logger::ERROR) {
            telegramError($message);
        }
    }

    public function updatePushNotificationsStats() {

        DB::beginTransaction();
        try {

            $pushNotificationsRead = request('push_read') != "" ? explode(",", request('push_read')) : [];
            $pushNotificationsOpened = request('push_opened') != "" ? explode(",", request('push_opened')) : [];

            foreach ($pushNotificationsRead as $pushId) {
                $push = PushNotificationSentUser::where('push_notification_id', '=', $pushId)->where('user', '=', $this->user->id)->first();
                if ($push && !$push->read) {
                    $push->read = 1;
                    $push->save();
                    $stats = PushNotificationSent::where('push_notification_id', '=', $pushId)->first();
                    $stats->increment('read');
                }
            }

            foreach ($pushNotificationsOpened as $pushId) {
                /** @var Model $push */
                $push = PushNotificationSentUser::where('push_notification_id', '=', $pushId)->where('user', '=', $this->user->id)->first();
                if ($push && !$push->opened) {

                    $wasRead = $push->read == 1;
                    $push->read = 1;
                    $push->opened = 1;
                    $push->save();
                    $stats = PushNotificationSent::where('push_notification_id', '=', $pushId)->first();
                    $stats->increment('opened');

                    if (!$wasRead) {
                        $stats->increment('read');
                    }

                }
            }

            DB::commit();
        } catch (\Exception $e) {
            DB::rollBack();
            $ex = new \Exception("Errore nell'aggiornamento delle notifiche push lette/aperte:\n".$e->getMessage(), $e->getCode());
            $this->onException($ex);
        }
    }

    public function addBonjourToHistory()
    {
        BonjourHistory::create(['user' => $this->user->id, 'device' => $this->device->id]);
    }

}
