<?php

namespace FourWords\AppManager\Ioc\Commands;

use FourWords\AppManager\Ioc\Database\Seeders\DatabaseSeeder;
use Illuminate\Console\Command;
use Illuminate\Contracts\Filesystem\FileNotFoundException;
use Illuminate\Filesystem\Filesystem;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Crypt;
use Illuminate\Support\Facades\File;
use Symfony\Component\Console\Input\InputOption;

class InstallCommand extends Command
{

    protected const NEEDED_EXTENSIONS = ["bcmath", "curl", "gd",
        "intl", "json", "mbstring", "mysqlnd",
        "PDO", "pdo_mysql", "pdo_sqlite", "mysqli",
        "xml", "xmlreader", "xmlwriter", "zip"
    ];

    /**
     * The console command name.
     *
     * @var string
     */
    protected $name = 'ioc:install';

    /**
     * The console command description.
     *
     * @var string
     */
    protected $description = 'Installa il Modulo iOC';

    /**
     * Metodo usato per specificare le opzioni del comando ioc:install
     * @return array|array[]
     */
    protected function getOptions()
    {
        return [
            ['force', null, InputOption::VALUE_NONE, 'Forza l\'operazione anche se già installato!', null]
        ];
    }

    /**
     * Execute the console command.
     *
     * @param Filesystem $filesystem
     *
     * @return void
     * @throws FileNotFoundException
     * @throws \Illuminate\Contracts\Container\BindingResolutionException
     */
    public function handle(Filesystem $filesystem)
    {

        // verifico se è già stato installato
        if (iOCIsInstalled()) {
            if ($this->option('force')) {
                File::delete(storage_path('.installed'));
            } else {
                $this->info('Il modulo iOC è già installato');
                return;
            }
        }

        // Verifico che le estensioni PHP necessarie siano installate
        if (($to_install = $this->missingExtensions()) != []) {
            $this->error('PHP extensions mancanti: ' . json_encode($to_install));
            return;
        }

        $this->call('optimize:clear');

        Artisan::call('vendor:publish --provider="Laravel\Sanctum\SanctumServiceProvider"');

        // Svuoto il file web.php con le rotte di default di Laravel
        $this->info('Rimuove rotte di default di Laravel in routes/web.php');
        $this->removeDefaultRoute($filesystem);

        // Genero le risorse iniziali nel database
//        (new DatabaseSeeder())->run();

        // Genero i dati demo
//        Artisan::call('ioc:demo');

        //Link storage
        $this->call('storage:link');

        //Sposto la migrazione per le queue se non esiste già (Abilitare nel .env QUEUE_CONNECTION=database
        $filteredArray = array_filter(glob(base_path('database/migrations/*')), function($element) {
            return strpos($element, 'create_jobs') !== false;
        });

        if (empty($filteredArray)) {
            $this->call('queue:table');
        }

        // Modifico package.json con le dipendenze JS, pubblico risorse, webpack.mix.js
        $this->call(UpdateCommand::class);

        // salvo un file per indicare la fine dell'installazione
        File::put(storage_path('.installed'), Crypt::encrypt(microtime()));

        // Fine installazione con successo!
        $this->info('Modulo iOC installato con successo!');
    }

    /**
     * Metodo che recupera le estensioni di php mancanti
     */
    protected function missingExtensions()
    {
        $array = array();
        foreach (self::NEEDED_EXTENSIONS as $ext) {
            if (!in_array($ext, get_loaded_extensions()))
                $array[] = $ext;
        }

        return $array;
    }

    /**
     * Metodo che rimuove la rotta di default da routes/web.php
     * @param $filesystem
     */
    protected function removeDefaultRoute($filesystem)
    {
        $routes_contents = $filesystem->get(base_path('routes/web.php'));
        $target = "Route::get('/', function () {";
        if (false !== strpos($routes_contents, $target)) {
            $contents = explode("\n", $routes_contents);
            $index = array_search($target, $contents);
            if ($index) {
                unset($contents[$index]);
                unset($contents[$index + 1]);
                unset($contents[$index + 2]);
            }
            $routes_contents = implode("\n", $contents);
            $filesystem->put(base_path('routes/web.php'), $routes_contents);
        }
    }

}
