<?php

namespace FourWords\AppManager\Ioc\Jobs;

use FourWords\AppManager\Ioc\Geo\ExceededLimitsException;
use FourWords\AppManager\Ioc\Geo\GeocoderServiceInterface;
use FourWords\AppManager\Ioc\Geo\GeocodingFailedException;
use FourWords\AppManager\Ioc\Geo\GeoUserData;
use FourWords\AppManager\Ioc\Geo\GeoUserFailed;
use FourWords\AppManager\Ioc\Geo\LocalGeocoder;
use FourWords\AppManager\Ioc\Geo\ThrottleException;
use FourWords\AppManager\Ioc\Models\IocUser;
use Monolog\Logger;

class GeocodeIocUsers extends IocJob
{

    protected $logLevel = Logger::INFO;

    protected $geocoded = [];

    public function handle()
    {
        // Imposto questo flag a false prima dell'handle della superclasse per evitare di riempire log se non ci sono invii da effettuare
        $this->shouldLog = false;
        parent::handle();
    }

    public function handleJob()
    {
        parent::handleJob();

        $q = IocUser::where(function ($query) {
            $query->whereNotNull('city')
                ->orWhereNotNull('cap')
                ->orWhereNotNull('province');
        });

        if (!empty($this->params['agente'])) {
            $q = $q->whereIn('agent', explode(',',$this->params['agente']));
        }

        if (!empty($this->params['codice-cliente'])) {
            $q = $q->whereIn('code', explode(',',$this->params['codice-cliente']));
        }

        if (!$forced = strtolower($this->params['forced'] ?? 'no') === "si") {
            $q = $q->doesntHave('geo');
        }

        $q = $q
            ->orderBy('active', 'desc')
            ->orderBy('created_at', 'desc');

        $usersToGeocode = $q->get();

        if (count($usersToGeocode) > 0) {
            if (!empty($this->params['agente'])) $this->logger->info("Filtro per agenti: ".$this->params['agente']);
            if (!empty($this->params['codice-cliente'])) $this->logger->info("Filtro per codici: ".$this->params['codice-cliente']);
            if ($forced) {
                $this->logger->info("Forzo geolocalizzazione per tutti gli utenti!");
            }
        } else {
            return;
        }

        $this->shouldLog = true;

        try {
            $geocoderName = basename(str_replace('\\', '/', trim(get_class(app()->get(GeocoderServiceInterface::class)), '\\')));
            $this->logger->info("Utilizzo servizio: ".$geocoderName);
        } catch (\Exception $ignored) {}

        /** @var IocUser $user */
        foreach ($usersToGeocode as $user) {

            // Genero l'hash che mi permette di evitare il geocoding ripetuto di indirizzi che falliscono
            $hash = md5($user->code.$user->province.$user->city.$user->cap.$user->address);

            if ($user->geoFailed()
                ->where('hash', $hash)
                ->exists()) {
                $this->logger->warning("Ignoro geocoding già fallito per utente ".$user->code);
                continue;
            }


            $geocoder = app()->make(GeocoderServiceInterface::class, [$user->cap, $user->province, $user->city, $user->address]);

            //TODO: Throttle preso da db, in base al geocoder (non fare throttling su localgeocoder)

            // Throttle max 5 richieste al secondo
            if (!$geocoder instanceof LocalGeocoder) usleep(200000); // 200,000 microsecondi / 0.2 secondi

            //TODO: Controllare di non aver superato i limiti giornalieri/mensili

            try {
                $geocodedData = $geocoder->geolocalize();
                if (!$geocodedData) {
                    throw new GeocodingFailedException("GeocodedData null");
                }

                $this->logger->info("Geocoded utente $user->code: ".json_encode($geocodedData, JSON_PRETTY_PRINT));

                $geoUserData = GeoUserData::withGeocodedData($geocodedData, $user);
                $geoUserData->save();

                GeoUserFailed::removeFor($user);

                $this->geocoded[] = $geoUserData;

            } catch (\Exception $e) {
                $this->logger->error("Impossibile geolocalizzare: " . json_encode($user->toArray(), JSON_PRETTY_PRINT));
                $this->logger->error($e->getMessage());

                if ($e instanceof ThrottleException) {
                    throw new ThrottleException("Errore, forse troppe richieste/s :: ".$e->getMessage(), $e->getCode());
                } else if ($e instanceof ExceededLimitsException) {
                    throw new ExceededLimitsException("Errore, forse esaurite richieste/mese :: ".$e->getMessage(), $e->getCode());
                } else if ($e instanceof GeocodingFailedException) {
                    // Se viene lanciata questa eccezione, non devo più ritentare il geocoding di quell'utente finché i dati rimangono uguali all'ultimo fallimento
                    $geoFailed = GeoUserFailed::for($user, $hash);
                    $geoFailed->save();
                }

            }

        }

        $this->logger->info("Geocodificati ".count($this->geocoded).'/'.count($usersToGeocode). " utenti");

    }

    public function parameters() {
        return [
            ['id' => 'codice-cliente', 'name' => 'Filtro su codici cliente',  'description' => '1234 oppure 1234, 5678, 9999'],
            ['id' => 'agente', 'name' => 'Filtro su agenti',  'description' => 'AGE oppure AGE, 123, 555'],
            ['id' => 'forced', 'name' => 'Forzato',  'description' => '"Si" se forzato']
        ];
    }
}
