<?php

namespace FourWords\AppManager\Ioc\Jobs;

use Carbon\Carbon;
use DateTimeZone;
use FourWords\AppManager\Ioc\Models\Task;
use FourWords\AppManager\Ioc\Models\TaskLog;
use Illuminate\Bus\Queueable;
use Illuminate\Contracts\Queue\ShouldBeUnique;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Queue\SerializesModels;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Str;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;

class IocJob implements ShouldQueue, ShouldBeUnique
{
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    public $timeout = 0;

    protected $logger;

    public $task;

    protected $error = null;

    protected $command;

    protected $params;

    protected $shouldLog = true;
    protected $logLevel = Logger::DEBUG;


    /**
     * Create a new job instance.
     *
     * @return void
     */
    public function __construct($task = null)
    {

        $this->connection = 'database';

        // Log
        $this->logger = new Logger(Str::kebab($this->getClassName()));
        $this->logger->setTimezone(new DateTimeZone('Europe/Rome'));

        if (!file_exists(config('ioc.log.dir'))) {
            mkdir(config('ioc.log.dir') . 'tasks/', config('ioc.log.dirPermissions'), true);
        }

        // Log generico del task, ruota ogni due settimane
        $LdebugHandler = new RotatingFileHandler(config('ioc.log.dir') . 'tasks/' . Str::kebab($this->getClassName()) . '.log', 14, Logger::DEBUG, true, config('ioc.log.filePermissions'));
        $debugDateFormat = "Y-m-d H:i:s";
        $debugOutput = "[%datetime%] %channel%.%level_name% : %message% \n";
        $debugFormatter = new LineFormatter($debugOutput, $debugDateFormat, true);
        $LdebugHandler->setFormatter($debugFormatter);
        $this->logger->pushHandler($LdebugHandler);

        if ($task != null) {
            $this->task = $task;
        } else {
            $this->task = Task::where('target_class', '=', get_class($this))->orWhere('target_class', '=', Str::replaceFirst('IocClient', 'Ioc', get_class($this)))->first();
        }

        if ($this->task == null) {
            throw new \Exception('Impossibile trovare il task ' . $this->getClassName());
        }


    }

    public function parameters()
    {
        return [];
    }

    protected function sanitizeParams() {
        $this->params  = array_filter($this->params , function ($value) {
            return $value !== "";
        });
    }

    public function setParameters($params)
    {

        $this->sanitizeParams();
        $this->params = $params;
    }

    public function getClassName()
    {
        return basename(str_replace('\\', '/', trim(get_class($this), '\\')));
    }

    public function setCommand($command)
    {
        $this->command = $command;
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle()
    {
        // Controllo che il job non sia stato schedulato troppo tempo prima che fosse eseguito dal queue:work
//        $jobRecord = $this->job->getJobRecord();
//
//        $createdAt = Carbon::createFromTimestamp($jobRecord->created_at/1000);
//        $reservedAt = Carbon::createFromTimestamp($jobRecord->reserved_at/1000);
//
//        if ($reservedAt->diffInMinutes($createdAt) >= 30) {
//            return;
//        }

        $joblog = app()->make(TaskLog::class);
        $joblog->task_id = $this->task->id;
        $joblog->start = time();
        $this->task->running = true;
        $this->task->save();

        // Log univoco di questo lancio
        $logFilePath = config('ioc.log.dir') . 'tasks/' . Str::kebab($this->getClassName()) . '-x.log';
        if (File::exists($logFilePath)) {
            File::put($logFilePath, '');
        }

        $LappHandler = new StreamHandler($logFilePath, $this->logLevel, true, config('ioc.log.filePermissions'));
        $LappHandler->setFormatter(new LineFormatter("[%datetime%] %channel%.%level_name% : %message% \n", "Y-m-d H:i:s", true));
        $this->logger->pushHandler($LappHandler);

        try {

            if ($this->shouldLog) {
                $this->logger->info('Inizio Task: ' . $this->task->name);
                if (!empty($this->params))
                    $this->logger->info('Parametri: ' . print_r($this->params, true));
            }

            $this->handleJob();

            $joblog->isOK = true;
        } catch (\Exception|\Throwable $e) {
            $this->logger->error('Errore Task: ' . $this->task->name);
            $this->logger->error($e->getMessage());
            $joblog->isOK = false;
            $joblog->error = $e->getMessage();
            $this->error = $e;
            if (app()->runningInConsole()) echo $e->getMessage();
            telegramError($e->getMessage() . "\nTASK IN ERRORE: " . $this->task->name);
        } finally {

            $joblog->end = time();

            if ($this->shouldLog) {
                $this->logger->info('Fine Task: ' . $this->task->name . ' - esito: ' . ($joblog->isOK ? 'POSITIVO' : 'ERRORE'));
            }

            if ($this->shouldLog || !$joblog->isOK)
                $joblog->save();

            $this->task->running = false;
            $this->task->save();

            File::copy($logFilePath, config('ioc.log.dir') . 'tasks/' . Str::kebab($this->getClassName()) . '-' . $joblog->start . '.log');

        }

        if ($this->error) {
            throw $this->error;
        }

    }

    public function handleJob()
    {
        //
    }

}
