<?php

namespace FourWords\AppManager\Ioc\Livewire\Table;

use FourWords\AppManager\Ioc\Traits\HasToasts;
use Illuminate\Pagination\LengthAwarePaginator;
use Illuminate\Support\Facades\Gate;
use Mediconesystems\LivewireDatatables\Http\Livewire\LivewireDatatable;


class IocTable extends LivewireDatatable
{
    use HasToasts;

    // Flag che sta ad indicare se valorizzare $allData con tutti i dati della query, più pesante rispetto a caricare
    // solo i dati paginati visibili, ma può essere utile ad es. per mostrare Marker su mappa
    protected $fetchAllData = false;
    protected $allData = [];

    public $exportable = true;
    public $hideable = 'select';
    public $dimOnLoading = true;

    public $allowUrlFilters = true;

    public $numberFilters = [];

    public function getResultsProperty()
    {
        $this->row = 1;

        if ($this->fetchAllData) {

            $realPage = $this->page;

            // Prendo 1000 risultati
            $res = $this->getQuery()->limit(1000)->get();

            // Li salvo in allData
            $this->allData = $res->toArray();

            $startIndex = ($realPage - 1) * intval($this->perPage);
            $endIndex = min($startIndex + intval($this->perPage) - 1, $res->count() - 1);

            $this->onAllDataFetched();

            // Passo il paginatore riciclando i risultati della query da 1000
            return $this->mapCallbacks(new LengthAwarePaginator(
                array_slice($this->allData,$startIndex, $endIndex - $startIndex + 1),
                $res->count(),
                intval($this->perPage),
                $realPage,
                ['path' => request()->url(), 'pageName' => 'page']
            ));
        }

        return $this->mapCallbacks(
            $this->getQuery()->paginate($this->perPage)
        );

    }

    protected function onAllDataFetched() {

    }

    public function setNumberFilterMinValue($index, $value)
    {
        $this->numberFilters[$index]['min'] = $value;
    }

    public function setNumberFilterMaxValue($index, $value)
    {
        $this->numberFilters[$index]['max'] = $value;
    }

    public function initialiseDefaultFilters()
    {
        if ($this->allowUrlFilters) {

            if ($queryFilters = request()->query() ) {
                $this->defaultFilters = $queryFilters;
//                $this->defaultFilters = array_merge($this->defaultFilters, );
            }
        }

        if (! $this->defaultFilters || ! is_array($this->defaultFilters) || count($this->defaultFilters) === 0) {
            return;
        }

        $columns = collect($this->columns);

        foreach ($this->defaultFilters as $columnName => $value) {
            $columnIndex = $columns->search(function ($column) use ($columnName) {
                return $column['filterName'] === $columnName;
            });

            if ($columnIndex === false) {
                continue;
            }

            $column = $columns[$columnIndex];

            if ($column['type'] === 'select' || (isset($column['filterable']) && is_array($column['filterable']))) {
                $this->activeSelectFilters[$columnIndex] = [$value];
            } else if ($column['type'] === 'string' || $column['type'] === 'editable') {
                $this->activeTextFilters[$columnIndex] = [$value];
//                $this->doTextFilter($columnIndex, $value);
            }

            if ($column['type'] === 'boolean') {
                $this->activeBooleanFilters[$columnIndex] = $value === "true" ? '1' : ($value === "false" ? '0' : $value);
            }


            if ($column['type'] === 'date') {
                $this->activeDateFilters[$columnIndex] = $value;
            }

            if ($column['type'] === 'datetime') {
                $this->activeDatetimeFilters[$columnIndex] = $value;
            }

            if ($column['type'] === 'time') {
                $this->activeTimeFilters[$columnIndex] = $value;
            }

            if ($column['type'] === 'number') {

                $filter = [];

                if (str_contains($value,'s')) {
                    preg_match('/s(\d+)/', $value, $matches);
                    if (isset($matches[1])) {
                        $f = intval($matches[1]);
                        $filter['start'] = $matches[1];
                        $this->setNumberFilterMinValue($columnIndex, $f);

                    }
                }

                if (str_contains($value,'e')) {
                    preg_match('/e(\d+)/', $value, $matches);
                    if (isset($matches[1])) {
                        $f = intval($matches[1]);
                        $filter['end'] = $matches[1];
                        $this->setNumberFilterMaxValue($columnIndex, $f);
                    }
                }

                $this->activeNumberFilters[$columnIndex] = $filter;


                $this->clearEmptyNumberFilter($columnIndex);
            }
        }
    }

    public function clearAllFilters()
    {
        foreach ($this->activeNumberFilters as $index => $value) {
            $this->setNumberFilterMinValue($index, '');
            $this->setNumberFilterMaxValue($index, '');
        }

        $this->activeDateFilters = [];
        $this->activeDatetimeFilters = [];
        $this->activeTimeFilters = [];
        $this->activeSelectFilters = [];
        $this->activeBooleanFilters = [];
        $this->activeTextFilters = [];
        $this->activeNumberFilters = [];
        $this->complexQuery = null;
        $this->userFilter = null;
        $this->visibleSelected = $this->selected;
        $this->setPage(1);
        $this->setSessionStoredFilters();

        $this->emitTo('complex-query', 'resetQuery');
    }

    protected function checkPermission($permission) {
        if(Gate::denies($permission)) {
            $this->errorToast('Non hai il permesso per questa azione');
            return false;
        }
        return true;
    }

    //TODO: Impostare permission a livello globale in ogni Table
    public function massDelete() {
        $this->model::destroy($this->selected);
        $this->selected = [];
        $this->visibleSelected = [];
    }

    public function shouldShowSelectAll() {
        return property_exists($this->getQuery()->first() ?? "", 'checkbox_attribute');
    }

    // Seleziono solamente i visibili nella pagina... override
    public function toggleSelectAll()
    {
        $allCheckboxes = array_map('strval', $this->getQuery()->get()->pluck('checkbox_attribute')->toArray());

        if (count($this->selected) >= intval($this->perPage)) {
            $this->selected = [];
        } else {
            $currentPage = $this->page;
            $perPage = $this->perPage;

            $start = ($currentPage - 1) * $perPage;
            $end = $start + $perPage;

            $this->selected = array_slice($allCheckboxes, $start, $end - $start);
        }

        $this->visibleSelected = $this->selected;
        $this->forgetComputed();
    }

    // Seleziono tutti gli elementi (anche non visibili)
    public function toggleAll() {
        if (count($this->selected) > 0) {
            $this->selected = [];
        } else {
            $allCheckboxes = array_map('strval', $this->getQuery()->get()->pluck('checkbox_attribute')->toArray());
            $this->selected = $allCheckboxes;
        }

        $this->visibleSelected = $this->selected;
        $this->forgetComputed();
    }

}
