<?php

namespace FourWords\AppManager\Ioc\Middleware;

use Carbon\Carbon;
use Closure;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Laravel\Sanctum\Events\TokenAuthenticated;
use Laravel\Sanctum\HasApiTokens;
use Laravel\Sanctum\Sanctum;

class IocApiMiddleware
{

    /**
     * The authentication factory implementation.
     *
     * @var \Illuminate\Contracts\Auth\Factory
     */
    protected $auth;

    /**
     * The number of minutes tokens should be allowed to remain valid.
     *
     * @var int
     */
    protected $expiration;

    /**
     * The provider name.
     *
     * @var string
     */
    protected $provider;

    public function unauthenticated() {
        return response()->json(['data' => ['now' => Carbon::now()->timestamp], 'status' => 'NO_AUTH','message' => 'Token non valido o scaduto'], 401);
    }

    public function getUserFromToken($token) {

        if (!$token) {
            return null;
        }

        $model = Sanctum::$personalAccessTokenModel;

        $accessToken = $model::findToken($token);

        if (! $this->isValidAccessToken($accessToken) ||
            ! $this->supportsTokens($accessToken->tokenable)) {
            return null;
        }

        $tokenable = $accessToken->tokenable->withAccessToken(
            $accessToken
        );

        event(new TokenAuthenticated($accessToken));

        if (method_exists($accessToken->getConnection(), 'hasModifiedRecords') &&
            method_exists($accessToken->getConnection(), 'setRecordModificationState')) {
            tap($accessToken->getConnection()->hasModifiedRecords(), function ($hasModifiedRecords) use ($accessToken) {
                $accessToken->forceFill(['last_used_at' => now()])->save();

                $accessToken->getConnection()->setRecordModificationState($hasModifiedRecords);
            });
        } else {
            $accessToken->forceFill(['last_used_at' => now()])->save();
        }


        return $tokenable;
    }

    /**
     * Handle an incoming request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  \Closure(\Illuminate\Http\Request): (\Illuminate\Http\Response|\Illuminate\Http\RedirectResponse)  $next
     * @return \Illuminate\Http\JsonResponse|\Illuminate\Http\RedirectResponse|\Illuminate\Http\Response
     */
    public function handle(Request $request, Closure $next)
    {

//        foreach (Arr::wrap(config('sanctum.guard', 'web')) as $guard) {
//            if ($user = $this->auth->guard($guard)->user()) {
//                return $this->supportsTokens($user)
//                    ? $user->withAccessToken(new TransientToken)
//                    : $user;
//            }
//        }

        if ( $tokenable = $this->getUserFromToken($this->getTokenFromRequest($request)) ) {
            Auth::login($tokenable);
            return $next($request);
        }

        return $this->unauthenticated();

    }

    /**
     * Determine if the tokenable model supports API tokens.
     *
     * @param  mixed  $tokenable
     * @return bool
     */
    protected function supportsTokens($tokenable = null)
    {
        return $tokenable && in_array(HasApiTokens::class, class_uses_recursive(
                get_class($tokenable)
            ));
    }

    /**
     * Get the token from the request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @return string|null
     */
    protected function getTokenFromRequest(Request $request)
    {
        if (is_callable(Sanctum::$accessTokenRetrievalCallback)) {
            return (string) (Sanctum::$accessTokenRetrievalCallback)($request);
        }

        return $request->bearerToken();
    }

    /**
     * Determine if the provided access token is valid.
     *
     * @param  mixed  $accessToken
     * @return bool
     */
    protected function isValidAccessToken($accessToken): bool
    {
        if (! $accessToken) {
            return false;
        }

        $isValid =
            (! $this->expiration || $accessToken->created_at->gt(now()->subMinutes($this->expiration)))
            && $this->hasValidProvider($accessToken->tokenable);

        if (is_callable(Sanctum::$accessTokenAuthenticationCallback)) {
            $isValid = (bool) (Sanctum::$accessTokenAuthenticationCallback)($accessToken, $isValid);
        }

        return $isValid;
    }

    /**
     * Determine if the tokenable model matches the provider's model type.
     *
     * @param  \Illuminate\Database\Eloquent\Model  $tokenable
     * @return bool
     */
    protected function hasValidProvider($tokenable)
    {
        if (is_null($this->provider)) {
            return true;
        }

        $model = config("auth.providers.{$this->provider}.model");

        return $tokenable instanceof $model;
    }

}