<?php
/**
 * Copyright (c) 2020/2
 * Created by Federico Papetti
 */

namespace FourWords\AppManager\Ioc\Providers;

use Carbon\Carbon;
use FourWords\AppManager\Director\FWUtils;
use FourWords\AppManager\Ioc\Bonjour\IocBonjour;
use FourWords\AppManager\Ioc\Commands\DumpAutoload;
use FourWords\AppManager\Ioc\Commands\InstallCommand;
use FourWords\AppManager\Ioc\Commands\ReorderPackages;
use FourWords\AppManager\Ioc\Commands\SeedDemoDataCommand;
use FourWords\AppManager\Ioc\Commands\UpdateCommand;
use FourWords\AppManager\Ioc\Contracts\Bonjour;
use FourWords\AppManager\Ioc\Controllers\AuthController;
use FourWords\AppManager\Ioc\Controllers\BonjourController;
use FourWords\AppManager\Ioc\Controllers\ResourceController;
use FourWords\AppManager\Ioc\Exceptions\IocExceptionHandler;
use FourWords\AppManager\Ioc\Geo\GeocoderServiceInterface;
use FourWords\AppManager\Ioc\Geo\LocalGeocoder;
use FourWords\AppManager\Ioc\Hashing\MD5Hasher;
use FourWords\AppManager\Ioc\Jobs\IocJob;
use FourWords\AppManager\Ioc\Middleware\HmacOrBearerMiddleware;
use FourWords\AppManager\Ioc\Middleware\IocApiMiddleware;
use FourWords\AppManager\Ioc\Middleware\LoggedIn;
use FourWords\AppManager\Ioc\Models\IocUser;
use FourWords\AppManager\Ioc\Models\Task;
use FourWords\AppManager\Ioc\Models\TaskLog;
use FourWords\AppManager\Ioc\Ps\PSBonjour;
use FourWords\AppManager\Ioc\Resource\ResourceRepository;
use FourWords\AppManager\Ioc\Telegram\iOCTelegramBot;
use Illuminate\Console\Events\CommandFinished;
use Illuminate\Console\Scheduling\Schedule;
use Illuminate\Contracts\Debug\ExceptionHandler as ExceptionHandlerContract;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Blade;
use Illuminate\Support\Facades\Cache;
use Illuminate\Support\Facades\Config;
use Illuminate\Support\Facades\Event;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Schema;
use Illuminate\Support\ServiceProvider;
use Illuminate\Support\Str;
use Livewire\Livewire;
class IocServiceProvider extends ServiceProvider
{

    public function register() {

        Config::set('app.timezone', 'Europe/Rome');
        date_default_timezone_set('Europe/Rome');

        director()->map4wPackages();
        $this->registerBindings();
        $this->registerPSBindings();
        $this->registerTasks();

        // Includo le configurazioni
        config(require __DIR__ . '/../../config/ioc.php');

        $this->app->bind(AuthController::class, function ($app, $params) {
            return new AuthController();
        });
    }

    public function boot()
    {

        set_time_limit(36000);
        error_reporting(E_ERROR);
        ini_set('memory_limit', '2048M');

        // Imposto l'asset_url di Livewire in caso laravel non sia nella root del dominio
        Config::set('livewire.asset_url', FWUtils::removeTrailingSlashes(asset('')));
        // Imposto il max size dei file temporanei in Livewire a 100MB
        Config::set('livewire.temporary_file_upload.rules', ['file', 'max:102400']);

        date_default_timezone_set('Europe/Rome');
        Config::set('app.timezone', 'Europe/Rome');
        app()->setLocale('it');
        Carbon::setLocale('it');

//        app()->make(IocExceptionHandler::class)->reportable(function (Throwable $e) {
//            if (app()->runningInConsole()) echo $e->getMessage();
//        });

        // Riordino i packages quando vengono eseguiti
        Event::listen(CommandFinished::class, function (CommandFinished $event) {
            if ($event->command === 'package:discover' || $event->command === 'config:clear' || $event->command === 'cache:clear') {
                Artisan::call('ioc:reorderpackages');
            }
        });

        // Cancello la cache quando eseguo una migrazione
        Event::listen(CommandFinished::class, function (CommandFinished $event) {
            if ($event->command === 'migrate') {
                Cache::clear();
            }
        });

        // Imposto l'hashing delle password con MD5
        app('hash')->extend('md5', function () {
            return new MD5Hasher();
        });
        Config::set('hashing.driver', 'md5');

        // Aggiungo il guard di Sanctum per l'autenticazione API di utenti iOC
        Config::set('auth.guards.sanctum', [
            'driver' => 'sanctum',
            'provider' => 'ioc_users'
        ]);

        // Aggiungo il guard per l'autenticazione web di utenti iOC
        Config::set('auth.guards.iocweb', [
            'driver' => 'session',
            'provider' => 'ioc_users'
        ]);

        // Aggiungo il provider di utenti iOC
        Config::set('auth.providers.ioc_users', [
            'driver' => 'eloquent',
            'model' => IocUser::class,
        ]);

        app('router')->aliasMiddleware('loggedin', LoggedIn::class);
        app('router')->aliasMiddleware('iocapi', IocApiMiddleware::class);
        app('router')->aliasMiddleware('hmacorbearer', HmacOrBearerMiddleware::class);

        $this->loadRoutesFrom(__DIR__ . '/../../routes/routes.php');
        $this->loadMigrationsFrom(__DIR__ . '/../../database/migrations');
        $this->loadViewsFrom(__DIR__ . '/../../resources/views', 'ioc');

        $this->registerConsoleCommands();
        $this->registerLivewireComponents();
        $this->registerBladeComponents();

        $this->addResourcePaths();

    }

    protected function addResourcePaths() {
        /** @var ResourceRepository $resourceRepository */
        $resourceRepository = $this->app->make(ResourceRepository::class);
        $resourceRepository->addType('img', [config('catalog.images.dir')]);
        $resourceRepository->addPath(config('catalog.resources.dir'), ['*']);
    }

    protected function registerConsoleCommands()
    {
        if ($this->app->runningInConsole()) {
            $this->commands(InstallCommand::class);
            $this->commands(UpdateCommand::class);
            $this->commands(ReorderPackages::class);
            $this->commands(DumpAutoload::class);
            $this->commands(SeedDemoDataCommand::class);
        }
    }

    // Registro tutti i componenti Livewire es. "GeoTable" -> "geo-table", "AdminCreateEditComponent" -> "admin-create-edit"
    protected function registerLivewireComponents()
    {
        if (app()->runningInConsole()) return;

        $filesDirectory = __DIR__.'/../Livewire';

        $livewireComponents = Cache::remember('livewire_components', app()->environment() == 'production' ? 86400 : 60, function () use ($filesDirectory) {

            $filePaths = [];
            foreach ( File::allFiles($filesDirectory) as $file) {
                $filePaths[] = ['realPath' => $file->getRealPath(), 'relativePath' => $file->getRelativePath()];
            }

            return $filePaths;
        });

        foreach ($livewireComponents as $livewireComponent) {

            $componentName = str_replace('.php', '', class_basename($livewireComponent['realPath']));

            $className = 'FourWords\\AppManager\\Ioc\\Livewire\\'. str_replace('/','\\',$livewireComponent['relativePath']). '\\' . $componentName;

            $componentName = str_replace('-component','',Str::kebab($componentName));

            Livewire::component($componentName, $className);
        }
    }

    protected function registerBladeComponents() {
        Blade::component('ioc::modal', 'modal');
        Blade::component('ioc::edit-modal', 'edit-modal');
    }

    protected function registerBindings() {

        $this->app->singleton(ExceptionHandlerContract::class, function () {
            return app()->make(IocExceptionHandler::class);
        });

        $this->app->singleton(IocExceptionHandler::class, function () {
            return new IocExceptionHandler($this->app);
        });

        $this->app->bind(iOCTelegramBot::class, function ($app, $params) {
            return new iOCTelegramBot();
        });

        $this->app->bind(Task::class, function ($app, $params) {
            return new Task();
        });

        $this->app->bind(TaskLog::class, function ($app, $params) {
            return new TaskLog();
        });

        $this->app->bind(IocJob::class, function ($app, $params) {
            return new IocJob(...$params);
        });

        $this->app->singleton(Bonjour::class, function () {
            return app()->make(IocBonjour::class);
        });

        $this->app->singleton(ResourceRepository::class, function () {
            return new ResourceRepository();
        });

        $this->app->bind('BonjourController', function ($app, $params) {
            return new BonjourController();
        });

        $this->app->bind('ResourceController', function ($app, $params) {
            return new ResourceController();
        });

        $this->app->bind(GeocoderServiceInterface::class, function ($app, $params) {
            return new LocalGeocoder(...$params);
        });

    }

    protected function registerPSBindings() {

        $this->app->bind(PSBonjour::class, function ($app) {
            return new PSBonjour();
        });

    }

    protected function registerTasks() {
        $this->app->booted(function () {

            try {
                if (!Schema::hasTable('ioc_tasks')) return;

                Task::all()->each(function ($task) {
                    if (class_exists(trim($task->target_class))) {

                        $this->app->bind($task->target_class, function ($app, $params) use ($task) {
                            return new $task->target_class(...$params);
                        });
                    }
                });
            } catch (\Exception $e) {
                Log::error($e->getMessage());
            }

        });
    }

    public static function scheduleTasks() {


            /** @var Schedule $schedule */
            $schedule = app()->make(Schedule::class);

            try {

                Task::all()->each(function ($task) use ($schedule) {
                    if (!$task->active) return;

                    if (class_exists($task->target_class)) {
                        /** @var IocJob $iocJob */
                        $iocJob = app()->make($task->target_class, [$task]);
                    } else {
                        /** @var IocJob $iocJob */
                        $iocJob = app()->make(IocJob::class, [$task]);
                    }

                    $schedule->job($iocJob)
                        ->timezone('Europe/Rome')
                        ->withoutOverlapping()  // Se volessi utilizzare diverse queue
                        ->{$task->frequency}($task->parameters)
                        ->runInBackground();

                });

            } catch (\Exception $e) {
                    Log::error($e->getMessage());
            }


    }


}
