<?php

namespace FourWords\AppManager\Orders;

use Illuminate\Support\Facades\Cache;
use Illuminate\Support\Facades\Http;

class B2B
{
    public $url;
    public $isGdprEnabled = false;
    public $appConfigPath;
    protected $appConfig;

    public $version; // semver

    public $release;

    protected $cacheTime = 21600; // 6 ore in secondi

    // costanti release b2b
    const B2B2010 = 'b2b2010';
    const B2B2011 = 'b2b2011';
    const B2B2012 = 'b2b2012';
    const B2BR12 = 'b2br12';
    const B2BR14 = 'b2br14';

    /**
     * release B2B in ordine
     *
     * @var array
     */
    public static $b2bReleases = [
        self::B2B2010,
        self::B2B2011,
        self::B2B2012,
        self::B2BR12,
        self::B2BR14,
    ];

    /**
     * lista codici pagamento attivi nel b2b
     * @var array
     */
    public $b2bPaymentCodes = array();

    /**
     * percorso della cartella contenente i documenti allegati alle new del b2b
     * @var string
     */
    public $b2bNewsFolderPath = '';

    /**
     * percorso della cartella contenente i documenti del b2b
     * @var string
     */
    public $b2bDocumentsFolderPath = '';

    /**
     * codice gruppo conto dei templates di creazione di un nuovi clienti
     * @var string
     */
    public $b2bTemplateNewCustomer = '';

    /**
     * codice del template di default per la creazione di un nuovo cliente
     * @var string
     */
    public $b2bNewCustomerCode = '';

    /**
     * flag mapping delle cartelle per login Name o codice gestionale
     * @var string [L,C]
     */
    public $b2bAgentsPrivateFolderType = '';

    /**
     * path di risposta per il completa ordine
     */
    public $b2bCompleteOrderPage = '';

    public function __construct($url = null, $appConfigPath = null, $version = null)
    {
        $this->url = config('orders.b2b.url');

        $this->appConfigPath = config('orders.b2b.appConfigPath');
        $this->isGdprEnabled = config('orders.b2b.isGdprEnabled') ?? Cache::get($this->url.'isGdprEnabled', false);
        $this->version = config('orders.b2b.version')  ?? $this->getVersion();

        if ($this->appConfigPath) {

            $this->appConfig = simplexml_load_file($this->appConfigPath);

            $this->_setRelease();
            $this->_setPaymentsCodes();
            $this->_setCompleteOrderPage();
            $this->_setDocsFolderPath();
            $this->_setTemplateOfNewCustomer();
            $this->_setNewCustomerCode();
            $this->_setNewsFolderPath();
            $this->_getAgentsPrivateFolderType();
        }

    }

    public function getVersion() {

        $this->version = Cache::get($this->url.'getVersion');
        if ($this->version) return $this->version;

        try {
            $response = Http::timeout(3)->get(config('orders.b2b.url') . 'getVersion');
        } catch (\Exception $e) {
            return null;
        }

        $result = utf8_encode($response->body());

        if ($response->status() != 200 || !$result) {
            return null;
        }

        $b2b_info = json_decode($result, true);

        //se lo stato della risposta è ok
        if ($b2b_info['status'] != 'OK') {
            return null;
        }

        $version = $b2b_info['data'];

        // verifico se la versione appartiene ad una di quelle con GDPR attivo
        // version > 3.0.1 or version >= 1.3.1
        if ($version['major'] >= 3) {
            $this->isGdprEnabled = (version_compare(sprintf("%s.%s.%s", $version['major'], $version['minor'], $version['modification']), '3.0.1', '>'));
        } elseif ($version['major'] >= 1) {
            $this->isGdprEnabled = (version_compare(sprintf("%s.%s.%s", $version['major'], $version['minor'], $version['modification']), '1.3.1', '>='));
        }

        $this->version = $version['major'] .".". $version['minor'] .".". $version['modification'];

        Cache::add($this->url.'isGdprEnabled',$this->isGdprEnabled, now()->addSeconds($this->cacheTime));
        Cache::add($this->url.'getVersion',$this->version, now()->addHours($this->cacheTime));

        return $this->version;
    }

    /**
     * metodo che confronta due release del b2b
     *
     * @param string $release1 release attuale
     * @param string $operator operatore (<,<=,==,=,!=,>=,>)
     * @param string $release2 release da confrontare
     *
     * @return boolean
     */
    public static function releaseCompare($release1, $operator, $release2): bool
    {

        $index1 = array_search($release1,self::$b2bReleases);
        $index2 = array_search($release2,self::$b2bReleases);

        switch ($operator) {
            case '<':  return $index1 < $index2;
            case '<=': return $index1 <= $index2;
            case '>':  return $index1 > $index2;
            case '>=': return $index1 >= $index2;
            case '==':
            case '=':  return $index1 == $index2;
            case '!=': return $index1 != $index2;
            default:   return false;
        }
    }

    /**
     * @throws \Exception
     */
    public function authenticate($params) {

        $url = config('orders.b2b.url');
        $params['user'] = strtoupper($params['user']);
        $params['pwd'] = $this->isGdprEnabled ? strtoupper($params['pwd']) : $params['pwd'];

        $response = Http::withOptions([
            'verify' => false,
            'timeout' => 300
        ])->get($url.'bonjour', $params);

        if ($response->status() !== 200) {

            switch ($response->status()) {
                case 404:
                    $errorMsg = $url." Impossibile interrogare il servizio";
                    break;
                case 500:
                    $errorMsg = $url. "Si è verificato un problema durante l'esecuzione del servizio";
                    break;
                default:
                    $errorMsg = strip_tags($response->body());
            }

            throw new \Exception($errorMsg);
        }

        return json_decode(utf8_encode($response->body()), true);
    }

    protected function _setRelease() {

        //recupero il percorso
        $release = (string)$this->appConfig->version['release'];

        //se è presente la versione la salvo nella variabile
        if($release != '') {
            $this->release = strtolower($release);
        }
    }

    /**
     * metodo che setta la lista dei codici pagamento del file xml
     */
    protected function _setPaymentsCodes() {

        //recupero dall'app-config quali codici recuperare dall'AS
        foreach ($this->appConfig->xpath('//tabs/tab[@name="CP"]/item') as $item) {
            $this->b2bPaymentCodes[] = (string)$item['key'];
        }
    }

    /**
     * metodo che setta il percorso della cartella contenente gli allegati delle news del b2b
     */
    protected function _setNewsFolderPath() {

        //recupero il percorso
        $path = (string)$this->appConfig->paths->news;

        //se la cartella esiste salvo il percorso nella variabile di classe
        if($path != '' && is_dir($path)) {
            $this->b2bNewsFolderPath = $path;
        }
    }

    /**
     * metodo che setta il percorso della cartella contenente i documenti del b2b
     */
    protected function _setDocsFolderPath() {

        //recupero il percorso
        $path = (string)$this->appConfig->paths->docs;

        //se la cartella esiste salvo il percorso nella variabile di classe
        if($path != '' && is_dir($path)) {
            $this->b2bDocumentsFolderPath = $path;
        }
    }

    /**
     * metodo che setta il gruppo conto e assegna il valore alla variabile di classe
     */
    protected function _setTemplateOfNewCustomer() {

        $tmp = $this->appConfig->xpath('//new-customer-prefs/template-code');
        $this->b2bTemplateNewCustomer = (string)$tmp[0];
    }

    /**
     * metodo che setta il codice del template di default da usare per la creazione di un nuovo cliente
     */
    protected function _setNewCustomerCode() {

        $tmp = $this->appConfig->xpath('//new-customer-prefs/default-template-code');
        $this->b2bNewCustomerCode = (string)$tmp[0];
    }

    /**
     * metodo che setta il flag che indica il metodo di creazione delle cartelle private nella cartella b2b dei documenti
     */
    protected function _getAgentsPrivateFolderType() {

        $tmp = $this->appConfig->xpath('//misc/download-folders-mapping');
        $this->b2bAgentsPrivateFolderType = (string)$tmp[0];
    }

    /**
     * metodo che setta il percorso della pagina del completa ordine del b2b
     */
    protected function _setCompleteOrderPage() {

        $tmp = $this->appConfig->xpath('//icm/complete-order-page');

        //estraggo il path
        $path = isset($tmp[0]) ? (string)$tmp[0] : '';

        //se non è presente
        if(empty($path) || $path == '') {

            //imposto uno di default
            $path = '/orders/finalSteps/cart.jsf';
        }

        //salvo nella variabile
        $this->b2bCompleteOrderPage = $path;
    }


}