<?php

namespace FourWords\AppManager\Orders\Services;

use Carbon\Carbon;
use DateTimeZone;
use FourWords\AppManager\Orders\Interfaces\OrderRepositoryInterface;
use FourWords\AppManager\Orders\Interfaces\OrderRowsRepositoryInterface;
use FourWords\AppManager\Orders\Interfaces\OrderServiceInterface;
use FourWords\AppManager\Orders\Model\AppModels\OrderResponse;
use FourWords\AppManager\Orders\Model\Order;
use Illuminate\Support\Facades\DB;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Logger;

abstract class AbstractOrderService implements OrderServiceInterface
{
    protected $orderRepository;
    protected $orderRowsRepository;
    protected $logger;

    /** @var object */
    protected $order;

    public function __construct()
    {

        // Log
        $this->logger = new Logger('orders');
        $this->logger->setTimezone(new DateTimeZone('Europe/Rome'));

        if (!file_exists(config('orders.log.dir'))) {
            mkdir(config('orders.log.dir'),config('orders.log.dirPermissions'), true);
        }

        // debug.log
        $LdebugHandler = new RotatingFileHandler(config('orders.log.dir').'debug.log',14,Logger::DEBUG, true, config('orders.log.filePermissions'));
        $debugDateFormat = "Y-m-d H:i:s";
        $debugOutput = "[%datetime%] %channel%.%level_name% : %message% \n";
        $debugFormatter = new LineFormatter($debugOutput, $debugDateFormat,true);
        $LdebugHandler->setFormatter($debugFormatter);
        $this->logger->pushHandler($LdebugHandler);
    }

    public function repository(): OrderRepositoryInterface
    {
        if (!$this->orderRepository) $this->orderRepository = app()->make(OrderRepositoryInterface::class);
        return $this->orderRepository;
    }

    public function rowsRepository(): OrderRowsRepositoryInterface
    {
        if (!$this->orderRowsRepository) $this->orderRowsRepository = app()->make(OrderRowsRepositoryInterface::class);
        return $this->orderRowsRepository;
    }

    /**
     * @throws \Throwable
     */
    public function send(object $order): object
    {
        $this->order = $order;

        $this->logger->debug("Dati ordine arrivati dall'app:");
        $this->logger->debug(json_encode($order, JSON_PRETTY_PRINT));
        $this->logger->debug("");

        $orderHistory = $this->repository()->find($order->dataCreazioneOrdine, $order->username);

        if (!$orderHistory) {
                $this->storeOrderLocally($order);
                $this->logger->error("Impossibile creare ordine con data $order->dataCreazioneOrdine e user $order->username");
            $this->logger->info("Creato nuovo ordine con data $order->dataCreazioneOrdine e user $order->username");
        } else {

            $this->logger->info("Recuperato ordine con data $order->dataCreazioneOrdine e user $order->username");

            if ($orderHistory->isSent()) {
                $this->logger->info("L'ordine è già stato inviato.");
                $order->response = new OrderResponse('OK', 'Ordine inviato correttamente', $order->cliente, $orderHistory->reference);
                return $order;
            }

            if ($orderHistory->hasErrors()) {
                $this->logger->info("L'ordine è in errore.");
            }

            $orderHistory->attempts += 1;
            $orderHistory->save();
        }

        try {
            $order->response = $this->storeRemotely($order);
        } catch (\Exception $e) {
            $this->logger->error("Invio remoto ordine ".$this->order->dataCreazioneOrdine." di ".$this->order->username." fallito con errore: \n".$e->getMessage());
            telegramError("Invio remoto ordine ".$this->order->dataCreazioneOrdine." di ".$this->order->username." fallito con errore: \n".$e->getMessage());

            $orderHistory->status = 'ERROR';
            $orderHistory->response_error = $e->getMessage();
            $orderHistory->save();

            $order->response = new OrderResponse('ERROR', 'Errore invio remoto ordine');
        }

        if ($order->response->status != 'OK') {
            $this->logger->error("Invio remoto fallito");
            return $order;
        }

        if (!$this->updateOrderLocally($order)) {
            $this->logger->error("Impossibile segnare l'ordine come inviato nel db locale, dopo l'invio remoto.");
            return $order;
        }

        $this->logger->info("Ordine inviato correttamente, stato: ".$order->response->status. " | Reference: ".$order->response->reference);

        return $order;

    }

    /**
     * @throws \Throwable
     */
    public function storeOrderLocally(object $order): bool
    {
        DB::beginTransaction();

        try {

            $orderRows = collect($order->rows);

            $orderTotal = $orderRows->map(function ($item) {
                return $item->qta * $item->prezzoNetto;
            })->sum();
            $quantityTotal = $orderRows->sum('qta');
            $colliTotal = $orderRows->sum('colli');

            $orderModel = $this->repository()->create([
                'creation_date' => $order->dataCreazioneOrdine,
                'username' => $order->username,
                'code' => $order->cliente,
                'agent' => $order->agente,
                'status' => Order::$WAIT,
                'total' => $orderTotal,
                'quantity_total' => $quantityTotal,
                'colli' => $colliTotal
            ]);

            if (!$orderModel) throw new \Exception("Impossibile salvare testata ordine");

            $insertedRows = $this->rowsRepository()->insertRowsFromApp($orderModel->id, array_values($order->rows));

            if (!$insertedRows) throw new \Exception("Impossibile salvare righe ordine");

            DB::commit();


        } catch (\Exception $e) {
            // If an exception occurs, rollback the transaction
            DB::rollback();
            throw $e;
        }

        return true;

    }

    public function updateOrderLocally(object $order): bool
    {
        return $this->repository()->update($order->dataCreazioneOrdine, $order->username, [
            'status' => $order->response->status,
            'reference' => $order->response->reference,
            'response_date' => Carbon::now(),
            'response_error' => $order->response->status !== Order::$OK ? $order->response->message : ""
        ]);
    }

}
