<?php

namespace FourWords\AppManager\PushNotifications\Livewire;

use Carbon\Carbon;
use FourWords\AppManager\Ioc\Helpers\TaskManager;
use FourWords\AppManager\Ioc\Livewire\IocComponent;
use FourWords\AppManager\PushNotifications\Jobs\SendPushNotifications;
use FourWords\AppManager\PushNotifications\Model\PushNotification;
use Illuminate\Support\Facades\Storage;
use Livewire\WithFileUploads;

class CreatePushNotificationPage extends IocComponent {
    use WithFileUploads;

    public $title;

    public
        $audience = [],
        $audienceSegment = [],
        $message = [],
        $delivery = [],
        $sendNow = false;

    public bool
        $androidSend = true,
        $iosSend = true;

    protected $listeners = ['setAudience', 'setAudienceSegment','setAudienceOS', 'setMessage', 'setDelivery'];

    protected $rules = [
        'title' => 'required|min:4',
    ];

    public function mount() {
        $this->initTitle();
    }

    public function render() {
        return view('push::livewire.create-page');
    }

    private function initTitle() {
        $notificationCount = PushNotification::all()->last()->id;
        $this->title = 'Nuova notifica '.($notificationCount+1);
    }
    
    public function create($toSend = true)
    {
        $notification = null;

        $this->validate();
        if(!$this->androidSend && !$this->iosSend) {
            $this->errorToast('Seleziona almeno un sistema operativo a cui inviare la notifica');
            return null;
        }

        try {
            $notification = $this->createNotificationDB($toSend);

            if ($this->sendNow) {
                TaskManager::runJob(SendPushNotifications::class);
            }

        } catch (\Exception $e) {
            $this->errorToast('Errore interno '. $e->getMessage());
            return null;
        }


        return ($notification != null && $this->sendNow) ? redirect(route('push.show',['id' => $notification->id])) : redirect(route('push.index'));
    }
    
    private function createNotificationDB($toSend) {
        $notification = new PushNotification();
        $notification->name = $this->title;

        $notification->title = $this->message['title'];
        //$notification->subtitle = $this->messageSubtitle;
        $notification->message = $this->message['description'];
        $notification->url = $this->message['url'];
        $notification->product = $this->message['productCode'];

        $notification->android_send = $this->androidSend;
        $notification->ios_send = $this->iosSend;

        $notification->draft = !$toSend;
        $notification->save();

        $notification->send_at = $this->getDateScheduled();

        $notification->image = $this->uploadImage($notification);
        $notification->save();

        $notification->groups()->attach(array_keys($this->audience));
        $notification->segments()->attach(array_keys($this->audienceSegment));

        return $notification;
    }

    private function uploadImage($table) {
        if($this->message['image'] != null) {
            $oldFilePath = 'public/notificationImages/'.$this->message['image'];
            $newFileName = 'notification-image'.$table->id.'.'.explode('.',$this->message['image'])[1];
            $newFilePath = 'public/notificationImages/'.$newFileName;

            Storage::move($oldFilePath, $newFilePath);
            return $newFileName;
        }
        return null;
    }

    public function setAudience($value) {
        $this->audience = $value;
    }

    public function setAudienceSegment($value) {
        $this->audienceSegment = $value;
    }

    public function setAudienceOS($android, $ios) {
        $this->androidSend = $android;
        $this->iosSend = $ios;
    }

    public function setMessage($value) {
        $this->message = $value;
    }

    public function setDelivery($value, $toSend, $sendNow) {
        $this->delivery = $value;
        $this->sendNow = $sendNow;
        $this->create($toSend);
    }

    private function getDateScheduled() {
        return Carbon::createFromFormat('Y-m-d H:i', $this->delivery['date'].' '.$this->delivery['hour'].':'.$this->delivery['minute'])->setTimezone('Europe/Rome');
    }
}
