<?php

/**
 * Copyright (c) 2020/2
 * Created by Federico Papetti
 */


namespace FourWords\AppManager\SqliteGen;

use FourWords\AppManager\Director\FWUtils;
use FourWords\AppManager\SqliteGen\Exceptions\EmptyModelException;
use Illuminate\Database\Schema\Blueprint;
use ReflectionClass;
use FourWords\AppManager\SqliteGen\Exceptions\ModelException;
use FourWords\AppManager\SqliteGen\Traits\Updatable;

/**
 * Class Model
 * @package SqliteGen
 *
 * Classe che deriva dal Model di Eloquent, sovrascrive alcune proprietà come il guarded, valorizza la proprietà table
 * e aggiunge i metodi per caricare dinamicamente le connessioni ed eseguire l'autoImport dei dati
 */
class Model extends \Illuminate\Database\Eloquent\Model
{


    /**
     * Trait custom per creare un model o aggiornare il dato se esistente
     *  <Model>::findOrCreate(<id>)
     */
    use Updatable;

    /** @var string Il nome della tabella nel database sqlite che riguarda questo model */
    public $sqliteTable;

//    /** @var string Flag che indica se la tabella è già stata creata con l'autoImport */
//    public $sqliteTableCreated = false;

    /** @var ReflectionClass */
    public $reflection;

    /** @var string Il database che si sta generando in questo momento */
    public $currentDatabase = '';

    /** @var string L'owner per cui si sta generando il database  */
    public $owner = '';

    public $timestamps = false;

    protected $guarded = [];

    protected $queryLoggerName = "";

    protected $useMutators = false;
    protected $logQueries = false;

    /** @var bool Lancio un'exception se non vengono restituiti dei risultati per popolare la tabella */
    protected $throwIfEmptyResult = false;

    /** @var SqliteGenerator $sqliteGenerator */
    protected $sqliteGenerator;

    public function __construct(array $attributes = [])
    {
        parent::__construct($attributes);

        $this->sqliteGenerator = app()->make('sqlitegen');
        $this->reflection = new ReflectionClass($this);

        // Aggiungo l'owner, in caso ne avessi bisogno nella query fatta con l'autoImport
        $this->owner = $this->sqliteGenerator->getCurrentOwner();
        $this->currentDatabase = $this->sqliteGenerator->getCurrentDatabase();

        /**  Aggiungo la connessione del relativo Database */
        // Creo dinamicamente la connessione del database di quest'entità, in caso non fosse già caricata.
        // Es. se sto generando solamente il database "specifico" che parte dal "generico", la connessione del generico non è caricata
        // ma se questo model che fa parte del generico viene chiamato dal db specifico, devo caricarla

        $connectionName = $this->getDatabaseName();
        $this->sqliteGenerator->loadConnection($connectionName);
        $this->setConnection($connectionName);

        // Imposto la tabella del Model se specificata
        if ($this->getSqliteTable()) {
            $this->setTable($this->getSqliteTable());
        }

    }

//    protected function usingExported() {
//        SQG::getExportedDBConnection();
//        $this->setConnection($this->getDatabaseName().'Exported');
//
//        return $this;
//    }

    public function getDatabaseName() {
        $arr = explode('\\', $this->reflection->getNamespaceName());
        $databaseName = $arr[count($arr)-2];
        return $databaseName;
    }

//    public static function createStatic() {
//        $instance = new static;
//
//          $arr = explode('\\', (new \ReflectionClass($instance))->getNamespaceName());
//          $count = count($arr);
//          $connectionName = $arr[$count-2];
//
//          $instance->setConnection($connectionName);
//
//        return $instance;
//    }
//
//    public static function __callStatic($method, $parameters)
//    {
//        return (new self)->$method(...$parameters);
//    }

//    /**
//     * @param array $record Record preso dal DataExtractor
//     * @return Model
//     */
//    public function autoMap($record) {
//
//        if (!is_array($record))
//            return $this;
//
//        foreach ($record as $key => $value) {
//            $this->{$key} = $value;
//        }
//
//        return $this;
//    }

    protected function up() {

        $this->getConnection()->getSchemaBuilder()->create($this->getSqliteTable(), function (Blueprint $table) {
            $this->autoImportUp($table);

            if (!$this->incrementing) {
//                $table->increments($this->primaryKey);
//            } else {
                $table->primary($this->primaryKey);
            }
        });

//        $this->sqliteTableCreated = true;
    }

    /**
     * @return string
     */
    protected function getSqliteTable() {
        return $this->sqliteTable;
    }

    protected function onInsertRows(&$rows) {
        return $rows;
    }

    protected function onInsertRecord(&$record) {
        return $record;
    }

    /**
     * Metodo per l'importazione automatica dei dati.
     * Nella classe figlia bisogna specificare lo schema in autoImportUp(), la query in autoImportQuery(), e autoImportConnection() se non si utilizza il dataextractor
     * @throws ModelException
     */
    public static function autoImport() {

        $instance = new static;

        $instance->queryLoggerName = $instance->owner != '' ? $instance->getConnectionName().$instance->owner : $instance->getConnectionName();

        $message = 'Utilizzo l\'autoImport per l\'entita '.$instance->reflection->getShortName().' con connessione '.$instance->autoImportConnection();
        $instance->sqliteGenerator->logger->info($message);

        $instance->logQuery($message);

        if (!$instance->getSqliteTable()) {
            throw new ModelException($instance,'Specificare getSqliteTable() per l\'import automatico dei dati');
        }

        if (!method_exists($instance, 'autoImportQuery')) {
            throw new ModelException('Specificare autoImportQuery() per l\'import automatico dei dati');
        }

        if (!$instance->sqliteGenerator->getConnection($instance->getConnectionName())->getSchemaBuilder()->hasTable($instance->sqliteTable)) {
            $instance->up();
        }

        $instance->logQuery('===============================');

        // Se la connessione ha il driver JDBC as400
        $config = config('database.connections.'.$instance->autoImportConnection());
        if ($config != null && $config['driver'] === 'com.ibm.as400.access.AS400JDBCDriver') {
            $instance->sqliteGenerator->getDataExtractor()->setConfig($instance->autoImportConnection());
            $rows = $instance->sqliteGenerator->getDataExtractor()->executeS($instance->autoImportQuery());
            $instance->logQuery('Query autoImport DataExtractor:');
        } else {
            // Altrimenti passo il nome della connessione al SqliteGenerator
            $instance->logQuery('Query autoImport:');
            $rows = $instance->sqliteGenerator->getConnection($instance->autoImportConnection())->select($instance->autoImportQuery());
        }

        $instance->logQuery($instance->autoImportQuery());
        $instance->logQuery('===============================');

        $instance->logQuery('==========================');
        $instance->logQuery('Esempio risultati query autoImport');
        $instance->logQuery(trim(json_encode((!empty($rows) ? $rows[0] : []))));
        $instance->logQuery('==========================');

        if ($instance->throwIfEmptyResult && count($rows) == 0) {
            throw new EmptyModelException($instance,"La query ha restituito un risultato vuoto.\nQuery:\n".$instance->autoImportQuery());
        }

        // Se nella autoImportQuery() lascio un SELECT * dal record tengo solamente
        // le colonne che sono anche nella tabella $sqliteTable per evitare errori con il model di Eloquent
        if (substr(str_replace("\n", '', str_replace(' ', '', $instance->autoimportQuery())), 0, 7) === 'SELECT*') {

            foreach ($rows as $key => $record) {

                $queryLogWasEnabled = false;
                if ($instance->sqliteGenerator->isQueryLogActive()) {
                    $queryLogWasEnabled = true;
                }

                if ($queryLogWasEnabled) {
                    $instance->sqliteGenerator->pauseQueryLog();
                }

                $tableColumns = $instance->getConnection()->getSchemaBuilder()->getColumnListing($instance->sqliteTable);
                $rows[$key] = array_intersect_key($record, array_flip($tableColumns));

                if ($queryLogWasEnabled) {
                    $instance->sqliteGenerator->resumeQueryLog();
                }

            }
        }

        $instance->insert($rows);

    }

    public static function populate($rows) {

        $instance = new static;

        $instance->queryLoggerName = $instance->owner != '' ? $instance->getConnectionName().$instance->owner : $instance->getConnectionName();

        $message = 'Popolo l\'entita '.$instance->reflection->getShortName();
        $instance->sqliteGenerator->logger->info($message);

        if (!$instance->logQueries) {
            $instance->sqliteGenerator->pauseQueryLog();
        } else {
            $instance->sqliteGenerator->resumeQueryLog();
        }

        $instance->logQuery($message);

        if (!$instance->getSqliteTable()) {
            throw new ModelException($instance,'Specificare getSqliteTable() per popolare i dati');
        }

        if (!$instance->sqliteGenerator->getConnection($instance->getConnectionName())->getSchemaBuilder()->hasTable($instance->sqliteTable)) {
            $instance->up();
        }

        $instance->logQuery('==========================');
        $instance->logQuery('Esempio dati per il popolamento');
        $instance->logQuery(trim(json_encode((!empty($rows) ? $rows[0] : []))));
        $instance->logQuery('==========================');

        $instance->insert($rows);

    }

    protected function insert($rows) {

        if (count($rows) == 0) {
            $this->logQuery('Array vuoto. Nessun insert');
            return;
        }

        $this->onInsertRows($rows);

        if ($this->useMutators) {
            foreach ($rows as $row) {

                // Trasformo le stringhe vuote '' in null
                $record = FWUtils::emptyStringToNull($row);

                /** @var Model $model */
                $model = $this->reflection->getName();
                $this->onInsertRecord($record);
                $instance = new $model($record);
                $instance->save();
            }
        } else {

//            INSERT INTO ... VALUES([...], [...], [...]) introdotto da Sqlite 3.7.11+ - più veloce

            $count = 0;
            $split = 500; // SQLITE_MAX_COMPOUND_SELECT non permette più di 500 insert contemporaneamente
            $start = 0;

            while ($count < count($rows)) {

                $values = [];

                $previousRecord = null;

                for ($index = $start; $index < count($rows); $index++) {

                    if ($rows[$index] == null) continue;

                    $rows[$index] = (array)$rows[$index];
                    $this->onInsertRecord($rows[$index]);


                    $rowWithoutNulls = FWUtils::removeNullsAndEmptyString($rows[$index]);

                    if ($previousRecord && (count($previousRecord) != count($rowWithoutNulls)) || array_diff_key($previousRecord, $rowWithoutNulls)) {

                        // Se il numero di colonne è diverso dal numero di colonne del record precedente, devo spezzare qui e ripartire da questo indice
                        $start = $index;
                        break;
                    }

                    $previousRecord = $rowWithoutNulls;

                    $rows[$index] = $rowWithoutNulls;
                    $rows[$index] = FWUtils::inQuotes($rows[$index]);


                    $values[] = "(" . implode(",", array_values($rows[$index])) . ")";
                    $count += 1;

                    if (count($values) == $split) {
                        $start = $index+1;
                        break;
                    }
                }

                $insert = "INSERT INTO ".$this->getSqliteTable()." (".implode(",",array_keys( $previousRecord)).") VALUES ".implode(",",$values);
                $this->getConnection()->unprepared($insert);
            }

        }

    }

    protected function getDatabase() {
        return $this->sqliteGenerator->getDatabase($this->getDatabaseName());
    }

    protected function shouldLogQueries() {
        return config('sqlitegen.log.force') || $this->getDatabase()->logQueries;
    }

    protected function logQuery($message) {
        if (!$this->shouldLogQueries()) {return;}
        $logger = $this->sqliteGenerator->getQueryLogger($this->queryLoggerName);
        if ($logger) {
            $logger->debug($message);
        }
    }

    /**
     * La connessione che viene utilizzata per l'importazione automatica dei dati.
     * Può fare riferimento ad una connessione del dataextractor o del modulo sqlitegen,
     * quindi viene cercata in config('database.connections.<nomeconnessione>')
     * @return string
     */
    protected function autoImportConnection() {
        return config('sqlitegen.autoImportConnection');
    }

}
