<?php

namespace FourWords\AppManager\SqliteGen\Parameters;

use FourWords\AppManager\SqliteGen\IocSqliteDB;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Log;

class ParametersRepository implements \ArrayAccess, ParametersInterface
{
    /**
     * All of the configuration items.
     *
     * @var array
     */
    protected $items = [];

    /** @var IocSqliteDB */
    protected $db;

    /**
     * Create a new configuration repository.
     *
     * @return void
     * @throws \Exception
     */
    public function __construct(IocSqliteDB $SQLiteDB = null)
    {
        $this->items = [];
        $this->setDb($SQLiteDB);

        $parameters = [];
        try {
            $parameters = $this->db->getConnection()->table('parameters')->select()->get()->toArray();
        } catch (\Exception $e) {
            Log::error("Impossibile recuperare tabella parameters nel db ".get_class($SQLiteDB));
        }
        foreach ($parameters as $index => $parameter) {

            if ($this->isJson($parameter->value)) {
                $parameter->value = json_decode($parameter->value);
            }

            Arr::set($this->items, $parameter->object, $parameter->value);
        }
    }

    /**
     * @return IocSqliteDB
     */
    public function getDb(): IocSqliteDB
    {
        return $this->db;
    }

    /**
     * @param IocSqliteDB $db
     */
    public function setDb(IocSqliteDB $db)
    {
        $this->db = $db;
    }

    /**
     * Determine if the given configuration value exists.
     *
     * @param  string  $key
     * @return bool
     */
    public function has($key)
    {
        return Arr::has($this->items, $key);
    }

    /**
     * Get the specified configuration value.
     *
     * @param  array|string  $key
     * @param  mixed   $default
     * @return mixed
     */
    public function get($key, $default = null)
    {
        if (is_array($key)) {
            return $this->getMany($key);
        }

        $value = Arr::get($this->items, $key, $default);

        if ($this->isJson($value)) {
            $value = json_decode($value);
        }

        return $value;
    }

    /**
     * Get many configuration values.
     *
     * @param  array  $keys
     * @return array
     */
    public function getMany($keys)
    {
        $config = [];

        foreach ($keys as $key => $default) {
            if (is_numeric($key)) {
                list($key, $default) = [$default, null];
            }

            $value = Arr::get($this->items, $key, $default);
            if ($this->isJson($value)) {
                $value = json_decode($value);
            }
            $config[$key] = $value;
        }

        return $config;
    }

    /**
     * Set a given configuration value.
     *
     * @param  array|string  $key
     * @param  mixed   $value
     * @return void
     */
    public function set($key, $value = null)
    {

        $keys = is_array($key) ? $key : [$key => $value];

        foreach ($keys as $key => $value) {

            Arr::set($this->items, $key, $value);

            if (is_object($value) || is_array($value) || $this->isJson($value)) {
                $value = json_encode($value);
            }

            $this->db->getConnection()->table('parameters')->updateOrInsert(['object' => $key],['value' => $value]);
        }

    }

    /**
     * Prepend a value onto an array configuration value.
     *
     * @param  string  $key
     * @param  mixed  $value
     * @return void
     */
    public function prepend($key, $value)
    {
        $array = $this->get($key);

        array_unshift($array, $value);

        $this->set($key, $array);
    }

    /**
     * Push a value onto an array configuration value.
     *
     * @param  string  $key
     * @param  mixed  $value
     * @return void
     */
    public function push($key, $value)
    {
        $array = $this->get($key);

        $array[] = $value;

        $this->set($key, $array);
    }

    /**
     * Get all of the configuration items for the application.
     *
     * @return array
     */
    public function all()
    {
        return $this->items;
    }

    /**
     * Determine if the given configuration option exists.
     *
     * @param  string  $key
     * @return bool
     */
    public function offsetExists($key)
    {
        return $this->has($key);
    }

    /**
     * Get a configuration option.
     *
     * @param  string  $key
     * @return mixed
     */
    public function offsetGet($key)
    {
        return $this->get($key);
    }

    /**
     * Set a configuration option.
     *
     * @param  string  $key
     * @param  mixed  $value
     * @return void
     */
    public function offsetSet($key, $value)
    {
        $this->set($key, $value);
    }

    /**
     * Unset a configuration option.
     *
     * @param  string  $key
     * @return void
     */
    public function offsetUnset($key)
    {
        $this->set($key, null);
    }

    function isJson($string) {
        return (!empty($string) && is_string($string) && ($string[0] == '{' || $string[0] == '['));
    }
}