<?php

/**
 * Copyright (c) 2020/2
 * Created by Federico Papetti
 */


namespace FourWords\AppManager\SqliteGen;

use DateTimeZone;
use Exception;
use FourWords\AppManager\Director\FWUtils;
use FourWords\AppManager\SqliteGen\Archive\ArchiverInterface;
use Illuminate\Database\Connection;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\File;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;
use Monolog\Processor\IntrospectionProcessor;
use Monolog\Processor\MemoryUsageProcessor;
use ReflectionClass;
use SQLite3;

use FourWords\AppManager\SqliteGen\Exceptions\HandledException;

/**
 * Class SQLiteDB
 * @package SqliteGen
 *
 * La classe che identifica il singolo SQLite, generico o per "owner"
 */
class SQLiteDB
{
    /** @var int L'ordine di esecuzione, un db con order 1 verrà eseguito prima di quello con order 5 etc...  */
    public $order = 1;

    /** @var string Il namespace o il nome del Database o il path completo di un database esterno
     * Se specificato verrà utilizzato come base e quindi la funzione up()
     * aggiungerà tabelle/colonne e dati nel db di partenza, che verrà poi popolato
     */
    public $startFrom;

    /**
     * @var bool
     * True se il database viene generato per più utenti e non è generico
     */
    public $hasOwners = true;

    /** @var string
     * Il valore che identifica il "proprietario" del database.
     * Questo valore verrà popolato ogni volta in base all'array ritornato da getOwners()
     * Può essere usato per estrarre dati solo per il proprietario del database, di conseguenza
     * verrà esportato in nella cartella exportDir()/$owner/nomedb.sqlite
     */
    protected $owner;

    /**
     * @var array
     * Array multidimensionale che specifica quali colonne escludere nella comparazione dei database sqlite
     * es. ['*' => ['created_at','updated_at'], 'dipendenti' => ['ts_badge_timbrato']]
     * Escludo le colonne created_at e updated_at in qualsiasi tabella
     * Escludo la colonna ts_badge_timbrato solo nella tabella dipendenti
     */
    public $columnsToNotCompare = ['*' => ['created_at','updated_at']];

    /** @var ReflectionClass */
    protected $reflection;

    /**
     * @var bool
     * Se impostato a true, il database non viene mai esportato ma viene solamente generato nella cartella temp.
     * Utile per evitare di esportare database di partenza, che vengono utilizzati solamente per creare db con owner
     */
    public $keepOnlyInTemp = false;

    /**
     * @var bool
     * Se impostato a true, il database viene esportato senza archivio
     */
    public $exportUncompressedDb = true;

    /**
     * @var bool
     * Se impostato a true, il database viene esportato anche come archivio
     */
    public $exportCompressedDb = false;

    /**
     * @var bool
     * Se impostato a true e se true anche exportCompressedDb il file compresso verrà versionato con il timestamp, stile icm
     */
    public $timestampVersioning = false;

    /**
     * @var int
     * Il numero di versioni da tenere, se il database viene zippato con la nomenclatura nome.timestamp.zip
     */
    public $versionsToKeep = 3;

    /**
     * @var bool
     * Se false non loggo le query di questa connessione
     */
    public $logQueries = false;

    /**
     * @var ArchiverInterface
     */
    protected $archiver;

    protected $cachedVersions = null;

    /** @var SqliteGenerator $sqliteGenerator */
    protected $sqliteGenerator;

    /**
     * SQLiteDB constructor.
     */
    public function __construct(SqliteGenerator $sqliteGenerator)
    {
        $this->reflection = new ReflectionClass($this);
        $this->sqliteGenerator = $sqliteGenerator;
    }

    /**
     * Creo il logger per le query del database
     * @return Logger
     */
    public function getQueryLogger() {

        if (!config('sqlitegen.log.force') && !$this->logQueries) return null;

        // Creo il logger che salverà le query di questo DB in log/NomeDB/query.log
        $log = new Logger($this->getName().'Query');

        $log->setTimezone(new DateTimeZone('Europe/Rome'));

        $log->pushProcessor(new IntrospectionProcessor());
        $log->pushProcessor(new MemoryUsageProcessor());

        if (!file_exists(config('sqlitegen.log.dir').$this->getName())) {
            mkdir(config('sqlitegen.log.dir').$this->getName(),config('sqlitegen.log.dirPermissions'), true);
        }

        $queryHandler = new RotatingFileHandler(config('sqlitegen.log.dir').$this->getName().'/query.log',14,Logger::DEBUG, true, config('sqlitegen.log.filePermissions'));
        $queryDateFormat = "Y-m-d H:i:s";
        $queryOutput = "[%datetime%] [%extra.memory_usage%] DB %context.db%: %message% \n";
        $queryFormatter = new LineFormatter($queryOutput, $queryDateFormat,true);
        $queryHandler->setFormatter($queryFormatter);
        $log->pushHandler($queryHandler);

        return $log;
    }

    /**
     * Recupero l'owner per cui sto creando il db
     * @return string
     */
    public function getOwner(): string
    {
        return $this->owner;
    }

    /**
     * Imposto l'owner per cui sto creando il db
     * @param string $owner
     */
    public function setOwner(string $owner)
    {
        $this->owner = $owner;
        $this->sqliteGenerator->removeQueryLogger($this->getName());
    }

    /**
     * Creo il logger per le query del database, specifico per l'owner
     * @return Logger
     */
    public function getOwnerQueryLogger() {

        if (!$this->logQueries) { return null; }

        // Creo il logger che salverà le query di questo DB in log/NomeDB/$owner/query.log
        $log = new Logger($this->getName().$this->owner.'Query');

        $log->setTimezone(new DateTimeZone('Europe/Rome'));

        $log->pushProcessor(new IntrospectionProcessor());
        $log->pushProcessor(new MemoryUsageProcessor());

        $queryHandler = new RotatingFileHandler(config('sqlitegen.log.dir').$this->getName().'/'.$this->owner.'/query.log',14,Logger::DEBUG, true, config('sqlitegen.log.filePermissions'));
        $queryDateFormat = "Y-m-d H:i:s";
        $queryOutput = "[%datetime%] [%extra.memory_usage%] DB %context.db%: %message% \n";
        $queryFormatter = new LineFormatter($queryOutput, $queryDateFormat,true);
        $queryHandler->setFormatter($queryFormatter);
        $log->pushHandler($queryHandler);

        return $log;
    }

    /**
     * Creo/ricreo il database o parto da uno esistente se specificato, creo la struttura
     * delle tabelle e faccio partire i seeder
     * @throws Exception
     */
    public function initiate() {

        $dbPath = $this->getDatabaseFullPath();

        if (!file_exists($this->getTempDatabasePath())) {
            if (!mkdir($this->getTempDatabasePath(),config('sqlitegen.databases.dirPermissions'),true)) {
                throw new Exception("Impossibile creare la cartella dove verranno salvati i file temporanei dei database sqlite da esportare");
            }
        }

        // Se non ho specificato un altro database da cui partire, ricreo il db
        if (!$this->startFrom) {

            if (file_exists($dbPath)) {
                $this->sqliteGenerator->logger->debug('Il db esisteva, lo cancello');
                unlink($dbPath);
            } else {
                $this->sqliteGenerator->logger->debug('Il db non esisteva, lo ricreo');
            }

            $db = new SQLite3($dbPath);
            $db->close();
            chmod($dbPath, config('sqlitegen.databases.filePermissions'));

        } else {
            $this->sqliteGenerator->logger->debug('Il db da cui partire è: '.$this->startFrom);

            // Se c'è un punto nella stringa, non è un namespace ma un path
            if (strpos($this->startFrom, ".") !== false) {
                $this->copyFrom($this->startFrom, true);
            } else {

                $lastBackslash = strrpos($this->startFrom, '\\');
                $dbName = substr($this->startFrom, ($lastBackslash ? $lastBackslash+1 : 0));
                $this->copyFrom($dbName);

                // Riapro la connessione al db, perché se avevo una precedente connessione al sqlite che è stato poi sostituito posso incorrere
                // nell'errore  General error: 17 database schema has changed
                $this->sqliteGenerator->getConnection($this->getName())->reconnect();
            }

        }

        // Creo le tabelle
        $this->sqliteGenerator->logger->debug('Creo le tabelle');
        $this->up();

        // Seedo le tabelle
//        $this->seed();

    }

    /**
     * Faccio partire il popolamento dei dati
     * @throws Exception
     */
    public function start() {

        // Faccio partire il popolamento dei dati
        $this->sqliteGenerator->logger->debug('Faccio partire il popolamento dei dati');
        $start = time();
        $this->populate();
        $this->sqliteGenerator->logger->info('Popolato in '.FWUtils::humanSeconds(time()-$start));

    }

    /**
     * Ottimizzo ed esporto il db, dopo aver controllato le differenze con il precedente
     * @throws Exception
     */
    public function finish()
    {

        // Ottimizzo il DB
        $this->sqliteGenerator->logger->debug('Eseguo il comando VACUUM');
        $start = time();
        $this->vacuum();
        $this->sqliteGenerator->logger->debug('VACUUM eseguito in '.FWUtils::humanSeconds(time()-$start));


//        // Chiudo la connessione
//        App::$log->debug('Chiudo la connessione '.$this->getConnection()->getName(). ' ('.$this->getConnection()->getDatabaseName().')');
//        $this->close();

        // Esporto il database nella cartella definita in exportDir() se ci sono dati diversi o altre condizioni
        if ($this->shouldExport()) {
            $this->sqliteGenerator->logger->notice('Lo esporto!');
            $this->export();
        } else {
            $this->sqliteGenerator->logger->notice('Non viene esportato');
        }
    }

    /**
     * Faccio partire i seeder delle tabelle se presenti - non utilizzato al momento
     */
    public function seed()
    {
        // Prendo i Seed e li ordino
        $seedsList = array_diff(scandir($this->getSeedsPath()), array('..', '.'));
        usort($seedsList, function ($a, $b) {
            return $a->order > $b->order;
        });

        if (count($seedsList) > 0) {
            $this->sqliteGenerator->logger->debug('Eseguo il seed delle tabelle');
        }

        // Li faccio partire
        foreach ($seedsList as $seedName) {
            $seedClass = $this->getFullSeedNamespace($seedName);
            if (class_exists($seedClass)) {
                $this->sqliteGenerator->logger->debug($seedClass);
                /** @var Seeder $seed */
                $seed = new $seedClass();
                $seed->run();
            }
        }

    }

    /**
     * Ottimizzo il db
     * @throws Exception
     */
    public function vacuum() {
        $this->getConnection()->statement('VACUUM');
    }

//    public function close() {
//        $this->getConnection()->disconnect();
//    }

    /********** Metodi da sovrascrivere eventualmente nella sottoclasse */

    /**
     * Il metodo che viene chiamata prima del populate, per creare le tabelle
     */
    protected function up() { }

    /**
     * Specificare le logiche con cui popolare i dati nelle tabelle del sqlite
     * @throws Exception
     */
    protected function populate() { }

//    protected function down() { }

    /**
     * Gli owner per cui verrà generato il database. Es. i "codici agente"
     * @return array
     */
    public function getOwners() { return []; }

    /**
     * Percorso in cui verrà esportato il database.
     * @return string
     */
    public function exportDir() { return config('sqlitegen.databases.exportDir'); }

    public function getNameOfExportedDatabase() {
        return $this->getName() . '.sqlite';
    }

    public function getNameOfExportedDatabaseWithoutExtension() {
        return $this->getName();
    }

    /********** */

    /**************** Esportazione e spostamenti */

    public function export()
    {
        $exportDir = FWUtils::addTrailingSlashes($this->exportDir());

        if ($this->owner) {
            $exportDir = FWUtils::addTrailingSlashes($exportDir.$this->owner);
        }

        if (!file_exists($exportDir)) {
            mkdir($exportDir, config('sqlitegen.databases.dirPermissions'), true);
        }

        $completeExportPath = $exportDir . $this->getNameOfExportedDatabase();

        if ($this->exportUncompressedDb) {
            copy($this->getDatabaseFullPath(), $completeExportPath);
            $this->sqliteGenerator->logger->debug('Esportato il db in ' . $completeExportPath);
        }

        $this->sqliteGenerator->logger->debug('Dimensioni database: '.FWUtils::humanFilesize(filesize($this->getDatabaseFullPath())));

        if ($this->exportCompressedDb) {
            $this->archiver = app()->make(ArchiverInterface::class);
            $this->archiver->addFile($this->getDatabaseFullPath(),$this->getNameOfExportedDatabase());
            $this->archiver->archive($exportDir . strtolower($this->getNameOfExportedDatabaseWithoutExtension()).($this->timestampVersioning ? '.'.time() : ''));
            $this->sqliteGenerator->logger->info('Compresso in '.$this->archiver->getCompressionTime(true)." da ".$this->archiver->getUncompressedSize(true)." a ".$this->archiver->getCompressedSize(true).". Riduzione del ".$this->archiver->getCompressionPercentage(true));
            $this->sqliteGenerator->logger->debug('Esportato il db compresso in ' . $this->archiver->getArchivePath());
        }

        // Se sto generando il database per più owner, avendolo appena esportato, cancello il file temporaneo per non lasciare file sporchi
        if ($this->owner) {
            if (file_exists($this->getDatabaseFullPath())) {
                File::delete($this->getDatabaseFullPath());
            }
        }

        // Rimuovo le vecchie versioni
        $this->removeOlderVersions();

    }

    /**
     * Copio il database sopra un altro database
     * @param $databaseName
     * @throws Exception
     */
    public function copyTo($databaseName)
    {
        copy($this->getDatabaseFullPath(),$this->sqliteGenerator->getDatabase($databaseName)->getDatabaseFullPath());
    }

    /**
     * Copio il database da un altro database, posso specificare
     * il nome del database o il path completo
     * @param string $database
     * @throws Exception
     */
    public function copyFrom($database, $isPath = false)
    {
        $databasePath = "";

        if ($isPath) {
            $databasePath = $database;
        } else {
            $databaseToCopy = $this->sqliteGenerator->getDatabase($database);
            $databasePath = $databaseToCopy->getDatabaseFullPath();
        }


        if (!file_exists($databasePath)) {
            throw new HandledException('Il database " ' . $database . ' " non esiste. Impossibile generare il database " ' . $this->getName() . ' "');
        }

        copy($databasePath, $this->getDatabaseFullPath());

    }

    /**************** Path e namespace */

    public function getFullSeedNamespace($seedName) {
        $seedName = str_replace('.php','',$seedName);
        return $this->reflection->getNamespaceName().'\\Seed\\'.$seedName;
    }

    public function getFullEntitiesNamespace($entityName) {
        $entityName = str_replace('.php','',$entityName);
        return $this->reflection->getNamespaceName().'\\Entities\\'.$entityName;
    }

    public function getDatabaseDirPath() {
        return config('sqlitegen.databases.tempDir').'/';
    }

    // Non utilizzato al momento
    public function getSeedsPath() {
        return config('sqlitegen.databases.tempDir').'/'.$this->getName().'/Seed/';
    }

    public function getTempDatabasePath() {
        if ($this->hasOwners) {
            return config('sqlitegen.databases.tempDir') . '/' . $this->getName() . '/'.$this->owner.'/';
        } else {
            return config('sqlitegen.databases.tempDir') . '/' . $this->getName() . '/';
        }
    }

    public function getDatabaseFullPath() {
        return $this->getTempDatabasePath().$this->getNameOfExportedDatabase();
    }

    public function getExportedDatabaseFullPath() {
        return FWUtils::addTrailingSlashes($this->exportDir()).$this->getNameOfExportedDatabase();
    }

    public function getLatestVersionedCompressedDb($onlyName = false, $cached = false) {
        $versions = $this->getVersions(true, true, $cached);

        if (count($versions) > 0) {
            return $onlyName ? Arr::last(explode('/',Arr::first($versions))) : Arr::first($versions);
        }

        return null;
    }

    public function getLatestVersionTimestamp($cached = false) {

        $versions = $this->getVersions(true, true, $cached);

        if (count($versions) > 0) {
            return Arr::first(array_keys($versions));
        }

        return null;

    }

    public function fileExists() {
        return file_exists($this->getDatabaseFullPath());
    }

    public function exportedFileExists() {
        return file_exists($this->getExportedDatabaseFullPath());
    }

    public function getName() {
        return $this->reflection->getShortName();
    }

    public function forOwner($owner) {
        $this->setOwner($owner);
        return $this;
    }

    /**************** Connessione */

    /**
     * Prendo la connessione di questo db o se specificata quella di un altro
     * @param string $name
     * @return Connection
     * @throws Exception
     */
    public function getConnection($name = null) {
        if (!$name) {
            return $this->sqliteGenerator->getConnection($this->getName());
        }
        // Se è stata passata una classe (es. Generic::class)
        if (class_exists($name)) {
            $name = last(explode('\\',$name)); // Prendo il nome dal namespace
        }
        return $this->sqliteGenerator->getConnection($name);
    }

    /**************** Comparazione */

    /**
     * Metodo che confronta questo sqlite con quello precedentemente esportato (se presente)
     *
     * @return boolean true se hanno almeno qualcosa di diverso, false se sono uguali
     * @throws Exception
     */
    public function shouldExport()
    {

        // Se è un database che utilizzo solo per creare altri database NON lo esporto, es. il generico
        if ($this->keepOnlyInTemp) {
            $this->sqliteGenerator->logger->notice('Il database deve rimanere in temp per la generazione di altri db');
            return false;
        }

        if ($this->sqliteGenerator->isForcingRelease()) {
            $this->sqliteGenerator->logger->debug('È richiesto il rilascio forzato');
            return true;
        }

        // Se non è mai stato esportato o è stato cancellato
        if($this->exportUncompressedDb && !$this->exportedFileExists()) {
            $this->sqliteGenerator->logger->notice('È il primo file');
            return true;
        }

        // Se non esporto il db "non compresso" vuol dire che non posso confrontare le tabelle quindi esporto forzatamente
        if (!$this->exportUncompressedDb) {
            $this->sqliteGenerator->logger->notice('Non posso confrontare le differenze perché esporto solo i file compressi');
            return true;
        }

        // Metto in pausa il logging delle query
        $this->sqliteGenerator->pauseQueryLog();

        $this->sqliteGenerator->logger->info('Controllo le differenze con il precedente');

//        $this->getConnection()->select('DETACH DATABASE "old"');

        // Eseguo l'attach del db che vorrei sovrascrivere, così da poter confrontare i dati
        try {
            $this->getConnection()->select('ATTACH "'.$this->getExportedDatabaseFullPath().'" as old');
        } catch (Exception $e) {
            $this->getConnection()->select('DETACH DATABASE "old"');
            $this->getConnection()->select('ATTACH "'.$this->getExportedDatabaseFullPath().'" as old');
        }

        // Ricavo tutti i nomi delle tabelle presenti nel db
        $tables_list = $this->getConnection()->table('sqlite_master')->select('tbl_name')->whereIn('type',['table','view'])->pluck('tbl_name')->toArray();

        // Per ogni tabella confronto i campi nel nuovo e nel vecchio db
        // nel caso dovessi trovare una differenza ritorno true
        foreach ($tables_list as $table_name)
        {

            // Prendo le tabelle che
            $columnsToIgnore = [];
            foreach ($this->columnsToNotCompare as $table => $columns) {
                if ($table == '*' || $table == $table_name) {
                    $columnsToIgnore = array_merge($columnsToIgnore,$columns);
                }
            }

            // Se la tabella non è presente nel vecchio db, sono diversi
            if($this->getConnection()->table('old.sqlite_master')->where([['type', '=', 'table'],['tbl_name','=',$table_name]])->count() == 0) {
                $this->sqliteGenerator->resumeQueryLog();
                return true;
            }

            // Se le due tabelle non hanno lo stesso numero di righe, i db sono diversi
            if($this->getConnection()->table($table_name)->count() != $this->getConnection()->table('old.'.$table_name)->count()) {
                $this->sqliteGenerator->resumeQueryLog();
                return true;
            }

            // Ricavo tutte le colonne di questa tabella presenti nel db, escludendo quelle specificate in $columnsToNotCompare
            $columns_list = array_diff(array_column($this->getConnection()->select('PRAGMA table_info("'.$table_name.'")'),'name'),$columnsToIgnore);

            // Ricavo tutte le colonne di questa tabella presenti nel db, escludendo quelle specificate in $columnsToNotCompare
            $columns_list_old = array_diff(array_column($this->getConnection()->select('PRAGMA old.table_info("'.$table_name.'")'),'name'),$columnsToIgnore);

            // Se le due tabelle hanno un numero diverso di colonne, i db sono diversi
            if(count($columns_list) != count($columns_list_old)) {
                $this->sqliteGenerator->resumeQueryLog();
                return true;
            }

            // Aggiungo parentesi quadre per evitare problemi con nomi di colonne riservati (es. "group", "rank"...)
            $columns_list_old = array_map(function ($str) { return "[$str]"; }, $columns_list_old);
            $columns_list = array_map(function ($str) { return "[$str]"; }, $columns_list);


            $query_table_old = "SELECT ".implode(',',$columns_list_old)." FROM old.$table_name";
            $query_table_new = "SELECT ".implode(',',$columns_list)." FROM $table_name";

            if( ((int) array_column($this->getConnection()->select("SELECT count(*) FROM ($query_table_new EXCEPT $query_table_old)"),'count(*)')[0]
                    + ((int) array_column($this->getConnection()->select("SELECT count(*) FROM ($query_table_old EXCEPT $query_table_new)"),'count(*)')[0])) > 0) {
                $this->sqliteGenerator->resumeQueryLog();
                return true;
            }

        }

        $this->sqliteGenerator->resumeQueryLog();
        return false;
    }

    public function getVersions($timestampAsKey = false, $descending = true, $cached = false) {

        if (!$cached || $this->cachedVersions == null) {
            //recupero tutti i db presenti nella cartella (dal più vecchio al più recente)
            $exportDir = FWUtils::addTrailingSlashes($this->exportDir());

            if ($this->owner) {
                $exportDir .= $this->owner . '/';
            }

            $dbList = array_merge(glob($exportDir . '*.*.zip'), glob($exportDir . '*.*.gz'));

            if (count($dbList) == 0) {
                return [];
            }

            // Mappo i database
            $databases = [];
            foreach ($dbList as $pos => $file) {

                //recupero il nome del file dal percorso
                $filename = explode('/', $file);
                $filename = end($filename);

                //suddivido le varie parti del nome del file
                list($name, $timestamp, $ext) = explode('.', $filename);

                $databases[$timestamp] = $file;

            }

            $this->cachedVersions = $databases;
        } else {
            $databases = $this->cachedVersions;
        }

        // Ordino per timestamp discendente
        krsort($databases, SORT_NUMERIC);

        return $timestampAsKey ? $databases : array_values($databases);

    }

    public function removeOlderVersions() {

        $db_list = $this->getVersions();

        // Se sono meno del minimo di versioni da tenere esco
        if(sizeof($db_list) <= $this->versionsToKeep) {
            return;
        }

        // Cancello quelli che superano $versionsToKeep
        $counter = 1;
        foreach ($db_list as $filename) {

            if ($counter > $this->versionsToKeep) {
                unlink($filename);
            }

            $counter++;
        }

    }

}
