<?php

/**
 * Copyright (c) 2020/2
 * Created by Federico Papetti
 */

namespace FourWords\AppManager\SqliteGen;

use DateTimeZone;
use Exception;
use FourWords\AppManager\DataExtractor\DataExtractor;
use FourWords\AppManager\Director\DatabaseManager;
use FourWords\AppManager\Director\FWUtils;
use FourWords\AppManager\Director\Processors\DiskSpaceProcessor;
use FourWords\AppManager\SqliteGen\Exceptions\EmptyModelException;
use Illuminate\Database\Connection;
use Illuminate\Events\Dispatcher;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\RotatingFileHandler;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;
use Monolog\Processor\IntrospectionProcessor;
use Monolog\Processor\MemoryUsageProcessor;
use Psr\Log\LoggerInterface;
use ReflectionClass;

use FourWords\AppManager\SqliteGen\Exceptions\HandledException;
use FourWords\AppManager\Ioc\Ps\Traits\PSConnection;


class SqliteGenerator
{

    /** @var ReflectionClass */
    protected $reflection;

    /** @var string[] */
    protected $databasesMapping = [];

    /** @var string[] */
    protected $databasesToLoad = [];

    /** @var SQLiteDB[] */
    protected $databases;

    /** @var SQLiteDB[] */
    protected $filteredDatabases;

    /** @var DataExtractor DataExtractor per query nel DB2 */
    protected $dataExtractor;

    /** @var [] Filtri per la generazione dei db */
    protected $filters = [];

    /** @var bool Forzo il rilascio dei database */
    protected $forcingRelease = false;

    /** @var string Nome del database che sto generando in questo momento */
    protected $currentDatabase = '';

    /** @var string Owner del database che sto generando in questo momento */
    protected $currentOwner = '';

    // LOG
    /** @var LoggerInterface Logger degli eventi */
    public $logger;

    /** @var LoggerInterface[] Logger delle query */
    protected $queryLoggers = [];

    /** @var bool Se false non loggo le query */
    protected $queryLogActive = true;

    protected $databasesToGenerate = 0;
    protected $databasesGenerated = 0;
    protected $databasesOwnersNotGenerated = [];

    /**
     * SQG constructor.
     */
    public function __construct()
    {

        // Istanzio i logger
        $this->instantiateLoggers();

        $this->reflection = new ReflectionClass(__CLASS__);

        // Mappo e carico le connessioni dei database sqlite in src/Databases
        $this->mapDatabases();
    }

    public function pushHandler($handler) {
        $this->logger->pushHandler($handler);
    }

    public function load(string $namespace, ...$namespaces) {
        $namespaces = array_merge([$namespace], $namespaces);
        $this->databasesToLoad = array_merge($this->databasesToLoad,$namespaces);

        $this->mapDatabases();
    }

    /**
     * Mappo tutti i SqliteDB passati tramite static::load(...)
     */
    protected function mapDatabases() {

        foreach ($this->databasesToLoad as $index => $namespace) {

            if (class_exists($namespace)) {

                /** @var SqliteDB $db */
                $db = new $namespace($this);
                $this->databases[$db->getName()] = $db;

                if ($logger = $db->getQueryLogger()) {
                    $this->addQueryLogger($logger,$db->getName()); //TODO:>> Inizializzare solamente per la generazione?
                }

            } else {
                $this->logger->error('Impossibile recuperare il SqliteDB "'. $namespace);
            }

        }

        if (!empty($this->databases)) {
            // Li ordino mantenendo l'associazione chiave-valore
            uasort($this->databases, function ($a, $b) {
                return $a->order > $b->order;
            });

        }

    }

    /**
     * Genero tutti i database (se $databases = null) o il database specificato, filtrabile con ->skipping e ->forOwners
     * @param null $databases
     * @return SqliteGenerator
     */
    public function generate($databases = null) {

        $this->logger->debug("Inizio a costruire i filtri");

        if ($databases != null) {

            if (is_string($databases)) {
                $databases = explode(",",str_replace(' ','', $databases));
            }

            $databases = (array) $databases;

            $dbs = [];
            foreach ($databases as $database) {
                $explodedNamespace = explode('\\', $database);
                $dbs[] = last($explodedNamespace);
            }
            $this->filters = ['db' => $dbs];
        }

        $this->logger->debug("Filtri applicati finora: " . json_encode($this->filters));
        return $this;
    }

    /**
     * Salto la generazione del/dei database passato/i
     * @param string|array|null $databases
     * @return SqliteGenerator
     */
    public function skipping($databases = null) {
        if ($databases != null) {

            if (is_string($databases)) {
                $databases = explode(",",str_replace(' ','', $databases));
            }

            $databases = (array) $databases;

            $skip = [];
            foreach ($databases as $database) {
                $explodedNamespace = explode('\\', $database);
                $skip[] = last($explodedNamespace);
            }
            $this->filters['skip'] = $skip;
        }

        $this->logger->debug("Filtri applicati finora: " . json_encode($this->filters));
        return $this;
    }

    /**
     * Forzo il rilascio dei database
     * @param $force
     * @return SqliteGenerator
     */
    public function forced($force) {
        $this->forcingRelease = $force;
        if ($force) {
            $this->logger->debug("!!! Forzo la rigenerazione dei database");
        }
        return $this;
    }

    /**
     * Genera il/i database per uno o più owner
     * @param string|array $owners
     * @throws HandledException
     */
    public function forOwners($owners = array()) {
        if ($owners != null) {

            if (is_string($owners)) {
                $owners = explode(",",str_replace(' ','', $owners));
            }

            $owners = (array) $owners;
            $this->filters['owner'] = $owners;
        }

        $this->logger->debug("Filtri applicati finora: " . json_encode($this->filters));
        $this->run();
    }

    /**
     * Genera il/i database per tutti gli owner
     * @throws HandledException
     */
    public function forAll() {
        $this->logger->debug("Filtri applicati finora: " . json_encode($this->filters));
        $this->run();
    }

    /**
     * Faccio partire la generazione di tutti i SQLiteDB in /Databases
     * @return bool
     * @throws HandledException
     */
    protected function run() {

        $this->logger->notice("===========================");
        $this->logger->notice("**** NUOVA GENERAZIONE ****");

        if (empty($this->databases)) throw new HandledException("La lista di database da generare è vuota, anche prima di applicare i filtri");

        // Se ho dei filtri per db li applico
        if (!empty($this->filters) && ((isset($this->filters['db'])) || isset($this->filters['skip']))) {

            // Filtro per database da generare
            if (isset($this->filters['db'])) {
                $this->filteredDatabases = array_intersect_key($this->databases, array_flip($this->filters['db']));
            }

            // Filtro per database da non generare
            if (isset($this->filters['skip'])) {
                $this->filteredDatabases = array_diff_key($this->filteredDatabases ?? $this->databases, array_flip($this->filters['skip']));
            }

        } else {
            // Altrimenti metto tutti i db nella lista dei filtrati
            $this->filteredDatabases = $this->databases;
        }

        $this->logger->debug("Database da generare: ".json_encode(array_keys($this->filteredDatabases)));
        if (empty($this->filteredDatabases)) {
            throw new HandledException("La lista di database da generare è vuota, filtri errati?");
        }

        // Li ciclo (solamente i filtrati)
        foreach ($this->filteredDatabases as $dbName => $database) {

            $this->currentDatabase = $dbName;
            $this->logger->notice('Elaboro il Database: '.$this->currentDatabase);

            if (FWUtils::hasTrait($database,PSConnection::class)) {
                $this->logger->debug('Il database utilizza l\'integrazione ICM');
            }

            /** SE ho specificato dei "proprietari" del database
             ** eseguo il popolamento e l'esportazione per ognuno di essi **/
            if ($database->hasOwners) {

                $databaseOwners = $database->getOwners();

                if (count($databaseOwners) == 0) {
                    throw new Exception("Il database ha hasOwners = true ma getOwners() ha restituito 0 risultati.");
                }

                // Applico i filtri, in caso stessi forzando solo determinati owner
                if ($this->filters != null && array_key_exists('owner',$this->filters)) {
                    $owners = $this->filters['owner'];
                } else {
                    $this->logger->debug('Prendo gli owner:');
                    $owners = $databaseOwners;
                }

                // Controllo che gli owners esistano nel getOwners del Database
                // (es. se imposto un owner tipo "abcde" non perde tempo a fare query con quell'owner perché non è nella query getOwners del database, quindi potenzialmente errato)
                $removedElements = array_diff($owners, $databaseOwners);
                $checkedOwners = array_intersect($owners, $databaseOwners);

                if (count($removedElements)) {
                    $this->logger->warning('Rimossi i seguenti owner (potrebbero essere errati): '.json_encode($removedElements));
                }

                if (count($checkedOwners) == 0) {
                    throw new Exception('Non ci sono owner per generare il database '.$this->currentDatabase);
                }

                $this->logger->notice("Owners: ".json_encode($checkedOwners, JSON_PRETTY_PRINT).(array_key_exists('owner',$this->filters) ? ' (imposti dai filtri)' : ' (da query)'));

                $this->databasesToGenerate = count($checkedOwners);
                $this->databasesGenerated = 0;
                $this->databasesOwnersNotGenerated = [];

                foreach ($checkedOwners as $owner) {
                    $database->setOwner(strval($owner));

                    if ($logger = $database->getOwnerQueryLogger()) {
                        $this->addQueryLogger($logger, $dbName.$owner);
                    }

                    $this->currentOwner = $owner;

                    // Aggiungo la connessione alla lista connessioni
                    $this->addConnection([
                        'driver'    => 'sqlite',
                        'database'  => $database->getDatabaseFullPath()
                    ], $dbName);

                    try {
                        $this->gen($database);
                        $this->databasesGenerated += 1;
                    } catch (EmptyModelException $e) {
                        $this->logger->error("Saltata generazione database con owner ".$this->currentOwner.". Errore:\n".$e->getMessage());
                        $this->databasesOwnersNotGenerated[] = $this->currentOwner;
                    }

                    $databasesNotGenerated = count($this->databasesOwnersNotGenerated);

                    $this->logger->notice('Generati '.$this->databasesGenerated.'/'.$this->databasesToGenerate.
                        ' - Mancanti: '.($this->databasesToGenerate - $this->databasesGenerated - $databasesNotGenerated).
                        ($databasesNotGenerated > 0 ? ' - In errore: '.$databasesNotGenerated : ''));
                    $this->logger->notice('------------------------------');

                }

            } else {
                $this->logger->notice("Il Database non ha owner e verrà generato solo una volta");

                // Aggiungo la connessione alla lista connessioni
                $this->addConnection([
                    'driver'    => 'sqlite',
                    'database'  => $database->getDatabaseFullPath()
                ], $dbName);

                $this->gen($database);

                $this->logger->notice('------------------------------');

            }

        }

        if (count($this->databasesOwnersNotGenerated) > 0) {
            $ownersNotGenerated = implode(", ",$this->databasesOwnersNotGenerated);
            $errorMessage = 'Database IN ERRORE: '.count($this->databasesOwnersNotGenerated).'/'.$this->databasesToGenerate."\nCodici: ".$ownersNotGenerated;
            $this->logger->notice($errorMessage);
            telegramError("Generazione database:\n".$errorMessage);
        }

        $this->logger->notice('**** GENERAZIONE CONCLUSA ****');
        $this->logger->notice('==============================');

        // Ho completato l'esecuzione
        return true;
    }

    /**
     * Generazione di un database
     * @param SQLiteDB $database
     */
    protected function gen(SQLiteDB $database) {
        $this->logger->notice('Genero il database '.$database->getName().($this->currentOwner != '' ? ' per l\'owner '.$this->currentOwner : ''));

        // Creo il db o parto da uno esistente, eseguo l'up() e il seed()
        $database->initiate();

        // Faccio partire il popolamento
        $database->start();

        // Eseguo il vacuum, controllo che il db non sia uguale al precedente, se non lo è lo esporto
        $database->finish();

    }

    /**
     * Ritorna true se sto forzando il rilascio dei db
     * @return bool
     */
    public function isForcingRelease() {
        return $this->forcingRelease;
    }

    /**
     * Ritorna l'istanza del DataExtractor
     * @return DataExtractor
     */
    public function getDataExtractor() {

        if (!$this->dataExtractor) {
            // Creo il DataExtractor
            $autoImportConnection = config('sqlitegen.autoImportConnection');
            $this->logger->debug("Creo il DataExtractor" . (config("database.connections.".$autoImportConnection) != null ? " con connessione ".$autoImportConnection : "- database.connections.".$autoImportConnection." non valorizzata"));
            $this->dataExtractor = new DataExtractor($autoImportConnection);
        }
        return $this->dataExtractor;
    }

    /**
     * Aggiungo la connessione SQLite al DatabaseManager
     * @param array $configurations
     * @param string $connectionName
     */
    public function addConnection($configurations, $connectionName) {
        $this->logger->debug('Aggiungo la connessione '.$connectionName. ' ('.$configurations['database'].')');

        // Aggiungo la connessione a Laravel
        DatabaseManager::addConnection($configurations,$connectionName);
        ;
        // Abilito il log delle query
        DatabaseManager::getConnection($connectionName)->enableQueryLog();
        DatabaseManager::getConnection($connectionName)->setEventDispatcher(new Dispatcher());
        DatabaseManager::getConnection($connectionName)->listen(function ($query) {
            $this->onQueryExecuted($query);
            $query->connection->flushQueryLog();
        });

    }

    /**
     * Recupero la connessione SQLite dal DatabaseManager o la creo dinamicamente
     * @param $connectionName
     * @return Connection
     * @throws Exception
     */
    public function getConnection($connectionName) {

        try {
            return DatabaseManager::getConnection($connectionName);
        } catch (Exception $e) {
            $this->addConnection(['driver' => 'sqlite', 'database' => $this->getDatabase($connectionName)->getDatabaseFullPath()], $connectionName);
        }

        return DatabaseManager::getConnection($connectionName);
    }

    /**
     * Carico la connessione dal Model.
     * I database potrebbero non essere ancora stati mappati oppure non esistere, in caso non esistano lancio un'eccezione poiché
     * non posso recuperare i suoi dati per fare le query dal Model
     * @param $connectionName
     * @throws Exception
     */
    public function loadConnection($connectionName) {

        // Se il database sqlite da cui voglio recuperare dei dati non è stato
        // ancora generato, lancio un'eccezione
        if (!$this->getDatabase($connectionName)->fileExists()) {
            throw new Exception('Il database '.$connectionName.' che sto richiedendo non è stato ancora generato.');
        }

        // Altrimenti carico la connessione se esiste o la aggiungo al volo
        try {
            // Se la connessione c'è già mi riconnetto, così da evitare problemi di cache se l'ho già utilizzata prima
            DatabaseManager::getConnection($connectionName)->reconnect();
        } catch (Exception $e) {
            $this->addConnection(['driver' => 'sqlite', 'database' => $this->getDatabase($connectionName)->getDatabaseFullPath()], $connectionName);
        }
    }

    /**
     * Recupero l'istanza SQLiteDB dal nome
     * @param $name
     * @return SQLiteDB
     * @throws Exception
     */
    public function getDatabase($name) {
        if (!array_key_exists($name, $this->databases)) {
            throw new HandledException("Il database " . $name . " non esiste");
        }
        return $this->databases[$name];
    }

    /**
     * Nome del database che sto generando in questo momento
     * @return string
     */
    public function getCurrentDatabase(): string
    {
        return $this->currentDatabase;
    }

    /**
     * Owner del database che sto generando in questo momento
     * @return string
     */
    public function getCurrentOwner(): string
    {
        return $this->currentOwner;
    }

    /**
     * Aggiungo un Logger alla lista di quelli che loggano le query
     * @param LoggerInterface $logger
     * @param string $connectionName
     */
    protected function addQueryLogger(LoggerInterface $logger, string $connectionName) {
        $this->queryLoggers[$connectionName] = $logger;
    }

    /**
     * Prendo un logger dalla lista di quelli che loggano le query
     * @param string $connectionName
     * @return LoggerInterface|null
     */
    public function getQueryLogger(string $connectionName) {
        if (array_key_exists($connectionName, $this->queryLoggers)) {
            return $this->queryLoggers[$connectionName];
        }

        return null;
    }

    /**
     * Rimuovo un logger dalla lista di quelli che loggano le query
     * @param string $connectionName
     */
    public function removeQueryLogger(string $connectionName) {
        if (array_key_exists($connectionName, $this->queryLoggers)) {
            unset($this->queryLoggers[$connectionName]);
        }
    }

    /**
     * Inizializzo il Logger principale con i suoi handler per app, debug... etc.
     */
    protected function instantiateLoggers() {

        $this->logger = new Logger('app');

        $this->logger->setTimezone(new DateTimeZone('Europe/Rome'));

        $this->logger->pushProcessor(new IntrospectionProcessor());
        $this->logger->pushProcessor(new MemoryUsageProcessor(false));
        $this->logger->pushProcessor(new DiskSpaceProcessor());

        if (!file_exists(config('sqlitegen.log.dir'))) {
            mkdir(config('sqlitegen.log.dir'),config('sqlitegen.log.dirPermissions'), true);
        }

        // app.log
        $handler = new RotatingFileHandler(config('sqlitegen.log.dir').'app/app.log',14,Logger::INFO, true, config('sqlitegen.log.filePermissions'));
        $dateFormat = "Y-m-d H:i:s";
        $output = "[%datetime%] [%extra.memory_usage%] %channel%.%level_name% %extra.function%: %message% \n";
        $formatter = new LineFormatter($output, $dateFormat,true);
        $handler->setFormatter($formatter);
        $this->logger->pushHandler($handler);

        // debug.log
//        $debugHandler = new RotatingFileHandler(config('sqlitegen.log.dir').'debug/debug.log',14,Logger::DEBUG, true, config('sqlitegen.log.filePermissions'));
//        $debugDateFormat = "Y-m-d H:i:s";
//        $debugOutput = "[%datetime%] [%extra.memory_usage%] [%extra.diskspace%] %channel%.%level_name% %extra.function%: %message% \n"; //%extra.class%::%extra.function%
//        $debugFormatter = new LineFormatter($debugOutput, $debugDateFormat,true);
//        $debugHandler->setFormatter($debugFormatter);
//        $this->logger->pushHandler($debugHandler);

        // error.log
        $errorHandler = new RotatingFileHandler(config('sqlitegen.log.dir').'error/error.log',14,Logger::ERROR, true, config('sqlitegen.log.filePermissions'));
        $errorDateFormat = "Y-m-d H:i:s";
        $errorOutput = "[%datetime%] [%extra.memory_usage%] %channel%.%level_name% %extra.class%::%extra.function%: %message% \n";
        $errorFormatter = new LineFormatter($errorOutput, $errorDateFormat,true);
        $errorHandler->setFormatter($errorFormatter);
        $this->logger->pushHandler($errorHandler);

    }

    /**
     * Metto in pausa il logging delle query
     */
    public function pauseQueryLog() {
        $this->queryLogActive = false;
    }

    /**
     * Riprendo in pausa il logging delle query
     */
    public function resumeQueryLog() {
        $this->queryLogActive = true;
    }

    public function isQueryLogActive() {
        return $this->queryLogActive;
    }

    /**
     * Metodo che viene eseguito dopo ogni query
     * @param $query
     */
    protected function onQueryExecuted($query) {

        if (!$this->queryLogActive) {
            return;
        }

        // Se ci sono dei binding (insert into ?,?...) li sostituisco con i valori
        $sql = FWUtils::replaceBindingsInQuery($query->sql, $query->bindings);

        // Loggo la query nella cartella di log del db in log/nomeDB/query/nomeDB.log
        $queryLogger = $this->getQueryLogger($query->connectionName);
        if ($queryLogger) {
            $queryLogger->debug($sql,['db' => $query->connectionName]);
        }

        // Loggo la query nella cartella di log specifica per l'owner del db in log/nomeDB/query/$owner/nomeDB.log
        $owner = $this->getCurrentOwner();
        if ($owner != '') {
            $queryLogger = $this->getQueryLogger($query->connectionName . $owner);
            if ($queryLogger) {
                $queryLogger->debug($sql, ['db' => $query->connectionName]);
            }
        }

    }

}
